#!/usr/bin/env python

'''
AMiGA library fo the microbial growth curve class.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"

# TABLE OF CONTENTS (2 functions and 1 class with 13 sub-functions)

# linearize
# maxValueArg
#
# Curve (CLASS)
# 	__init__
#   compute_mse
#   log_to_linear
#   describe
#   AreaUnderCurve
#   CarryingCapacity
#   MaxGrowthRate
#   MinGrowthRate
#   StationaryDelta
#   LagTime
#   data
#   sample
#   plot

import numpy as np # type: ignore
import pandas as pd # type: ignore
import matplotlib.pyplot as plt # type: ignore

from scipy.stats import norm # type: ignore

from .diauxie import detectDiauxie
from .utils import getValue


def linearize(arr,baseline,floor=True,logged=True):
    '''
    Converts arr from log space to linear space.

    Args:
        arr (numpy.ndarray): typically log OD measurmenets over time
        baseline (float): should be baseline OD at first time point
        floor (float): if you want to shift curves to start at zero, pass True, else False

    Returns:
        self.predicted_OD (numpy.ndarray): same shape as input
    '''

    if arr is None:
        return None

    # add ln OD(0) and exponentiate
    if logged:
        arr = np.exp(arr+np.log(baseline))
    elif not logged:
        arr = arr + baseline

    # subtract OD(0)
    if floor:
        arr = arr - arr[0] 

    return arr


def maxValueArg(x,y):
    '''
    Find the maximum value (of y) and argument (of x) that maximizes it.

    Args:
    	x (np.ndarray)
    	y (np.ndarray)

    Returns:
        maxValue (float): x value corresponding to maximum value of y
        maxArg (float): maximum value of y
    '''

    assert x.shape[0] == y.shape[0]
    assert x.shape[1] == 1
    assert y.shape[1] == 1

    ind = np.where(y==y.max())[0][0]
    maxValue = y[ind,0]
    maxArg = float(x[ind][0])

    return maxValue, maxArg


class GrowthCurve:

    def __init__(self,x=None,y=None,y0=None,y1=None,y2=None,cov0=None,cov1=None,baseline=1.0,name=None,logged=True):
        '''
        Data structure for a growth curves. This is primarily used for computing 
            growth curve parameters and characteistics or converting curve and its fit to 
            real space. 

        Args:
            x (numpy.ndarray): independent variable N x D, where N is number of measurements, and D is number of dimesions
            y (nump.ndaray): dependent variable Nx1, typically log OD (input to GP model)
            y0 (numpy.ndarray): dependent variable N x 1, typically log OD (outupt of GP model)
            y1 (numpy.ndarray): first-order derivative of y0, typically d/dt log OD
            y2 (numpy.ndarray): second-order derivative of y1, typically d^2/dt^2 log OD
            cov0 (numpy.ndarray): covariance matrix for dependent variable y0
            cov1 (numpy.ndarray): covariance matrix for first-order derivatie y1
            baseline (float): the OD at time zero for dependent variable, y
        '''

        # verify data types
        for arg,value in [('x',x),('y0',y0),('y1',y1)]:
            assert isinstance(value,np.ndarray), f"{arg} must be a numpy ndaray"

        assert x.shape[1] == 1, "x must be one-dimensional"
        assert y0.shape == y1.shape, "y0 and y1 must have the same shape"

        if y2 is not None:
            assert isinstance(y2,np.ndarray), "y2 must be a numpy ndarray"
            assert y0.shape == y1.shape == y2.shape, "y2 must have the same shape as y0 and y1"

        # define attributes
        self.name = name
        self.baseline = baseline 
        self.logged = logged
        self.x = x
        self.y = y
        self.y0 = y0
        self.y1 = y1
        self.y2 = y2
        self.cov0 = cov0
        self.cov1 = cov1

        # derive linear transformations of select attributes
        self.log_to_linear()

        # compute all growth parameters or characteristics
        self.describe()

    def compute_mse(self,pooled=False):
        '''
        Computes Mean Squared Error
        '''

        if pooled:
            self_data = self.data()
            y = self_data.GP_Input.values
            y_hat = [ii for ii in self_data.GP_Output.values if ~np.isnan(ii)]
            y_hat = y_hat * int(len(y)/len(y_hat))
        else:
            y = self.linear_input_raised
            y_hat = self.linear_output_raised

        mse = (1./y.shape[0]) * sum((y-y_hat)**2)

        return mse


    def log_to_linear(self):
        '''
        Converts actual and predicted log OD from log space to linear space.
        '''

        self.linear_input = linearize(self.y,baseline=self.baseline,floor=True,logged=self.logged)
        self.linear_input_raised = linearize(self.y,baseline=self.baseline,floor=False,logged=self.logged)
        self.linear_output = linearize(self.y0,baseline=self.baseline,floor=True,logged=self.logged)
        self.linear_output_raised = linearize(self.y0,baseline=self.baseline,floor=False,logged=self.logged)


    def describe(self):

        dx_ratio_min = getValue('diauxie_ratio_min')
        dx_ratio_varb = getValue('diauxie_ratio_varb')

        self.AreaUnderCurve()
        self.CarryingCapacity()
        self.MaxGrowthRate()
        self.MinGrowthRate()
        self.LagTime()
        self.StationaryDelta()
        
        params = {'auc_lin':self.auc_lin,
                  'auc_log':self.auc_log,
                  'k_lin':self.K_lin,
                  'k_log':self.K_log,
                  't_k':self.t_K,
                  'gr':self.gr,
                  'dr':self.dr,
                  'td':self.td,
                  't_gr':self.t_gr,
                  't_dr':self.t_dr,
                  'death_lin':self.death_lin,
                  'death_log':self.death_log,
                  'lagC':self.lagC,
                  'lagP':self.lagP}

        if self.y2 is not None: 

            dx = detectDiauxie(self.x,self.y0,self.y1,self.y2,self.cov0,self.cov1,
                              thresh=dx_ratio_min,varb=dx_ratio_varb)

            # describe all phases
            df_dx = []
            for idx,row in dx.iterrows():
                t0,t1 = row['t_left'],row['t_right'] # indices
                t0,t1 = (np.where(self.x==ii)[0][0] for ii in [t0,t1]) # time at indices
                if (t0 == 0) and (t1==(len(self.x)-1)):
                    dx_params = params
                    dx_params['t0'] = row['t_left']
                    dx_params['tf'] = row['t_right']
                    df_dx.append(pd.DataFrame(dx_params,index=[idx]))
                else:
                    curve = GrowthCurve(x=self.x[t0:t1],
                                        y0=self.y0[t0:t1]-self.y0[t0],
                                        y1=self.y1[t0:t1],
                                        cov0=self.cov0[t0:t1,t0:t1],
                                        cov1=self.cov1[t0:t1,t0:t1])
                    dx_params = curve.params
                    dx_params['t0'] = row['t_left']
                    dx_params['tf'] = row['t_right']
                    df_dx.append(pd.DataFrame(dx_params,index=[idx]))

            df_dx = pd.concat(df_dx,axis=0)
            df_dx.columns = [f'dx_{ii}' for ii in df_dx.columns]

            params.update({'diauxie':[1 if dx.shape[0] > 1 else 0][0],'df_dx':df_dx})

        self.params = params


    def AreaUnderCurve(self):
        '''
        Computes the Area Under the Curve (AUC).
        '''

        dt = np.mean(self.x[1:,0]-self.x[:-1,0])  # time interval (int)

        mu_Log = self.y0 # get log or linear OD
        mu_Lin = self.linear_output # get log or linear OD

        D_Log = np.repeat(dt,mu_Log.shape[0]).T  # area under each interval (np.ndarray), size is (x.shape[0],)
        D_Lin = np.repeat(dt,mu_Lin.shape[0]).T  # area under each interval (np.ndarray), size is (x.shape[0],)        

        self.auc_lin = np.dot(D_Lin,mu_Lin)[0] # cumulative sum of areas
        self.auc_log = np.dot(D_Log,mu_Log)[0] # cumulative sum of areas


    def CarryingCapacity(self):
        '''
        Computes the maximum carrying capacity.
        '''    

        #print(self.x)
        self.K_log, self.t_K = maxValueArg(self.x,self.y0)
        self.K_lin, self.t_K = maxValueArg(self.x,self.linear_output)


    def MaxGrowthRate(self):
        '''
        Computes the maximum specific growth rate (gr) and generation doubling Time (td).
        '''

        self.gr, self.t_gr = maxValueArg(self.x,self.y1)

        # compu
        if self.gr == 0:
            self.td = np.inf
        else:
            self.td = (np.log(2.0))/self.gr


    def MinGrowthRate(self,after_max=True):
        '''
        Computes the minimum of the derivative, which will often be the maximum death rate (dr). 
        '''

        x_K = int(np.where(self.x[:,0]==self.t_K )[0]) # index (not time) at maximum growth

        mu = self.y1
        if after_max:
            mu = mu[x_K:,0]

        x_dr = np.where(mu==mu.min())[0][0]
        t_dr = float(self.x[x_dr+x_K][0])

        minGrowthRate = mu[x_dr]

        self.dr, self.t_dr = minGrowthRate, t_dr


    def StationaryDelta(self):
        '''
        Computes difference between carrying capcaity and final OD.
        '''
        
        if self.K_lin <= 0:
            self.death_log = 0
            self.death_lin = 0
        else: 
            self.death_log = np.abs(self.y0[-1][0] - self.K_log)
            self.death_lin = np.abs(self.linear_output[-1][0] - self.K_lin)


    def LagTime(self):
        '''
        Computes the lag time either the classical definition or a probabilistic definition.
            The former defines the lag time as the intersection with the axis parallel to time 
            of the tangent intersecting the derivative of the latent function at maximum growth. 
            This tangent has slope m equivalent to the maximum of the derivative of the latent.
            The latter defines lag time as the time at which the 95-percent credible interval of  
            the growth rate (i.e. derivative of latent) deviates from zero. 

        Args:
            mode (str): either 'Classical' or 'Probabilistic
            threshold (float): Confidence Interval, used for probabilistic inference of lag time.
        '''

        x = self.x
        y0 = self.y0
        y1 = self.y1
        cov1 = self.cov1

        # CLASSICAL MODE

        t_gr = self.t_gr # time at maximal growth rate
        x_gr = int(np.where(x[:,0]==t_gr)[0]) # index at maximal growth rate

        m1 = y1[x_gr] # slope at maximal growth rate
        m0 = y0[x_gr] # log OD at maximal growth rate

        if m1 == 0:
            lagC = np.inf # no growth, then infinite lag
        else:
            lagC = (t_gr - (m0/m1))[0]

        # PROBABILISTIC MODE

        confidence = getValue('confidence_adapt_time')

        prob = np.array([norm.cdf(0,m,np.sqrt(v)) for m,v in zip(y1[:,0],np.diag(cov1))])

        ind = 0
        while (ind < prob.shape[0]) and (prob[ind] > confidence):
            ind += 1

        if ind == prob.shape[0]:
            lagP = np.inf
        else:
            lagP = float(self.x[ind][0])

        self.lagC = lagC
        self.lagP = lagP


    def data(self):
        '''
        Summarizes the object's data, including estimates of best fit for data and its derivative using GPs.

        Args:
            sample_id (varies): can possibly be float, int, or str
        Returns:
            df (pandas.DataFrame): each row is a specific time-point. Identifying columns is Time (and possibly Sample_ID).
                Data columns include
                + OD_Fit (gp model fit of original data)
                + OD_Derivative (gp model fit of derivative, insensitive to y-value, i.e. whether OD is centered)
                + GP_Input (input to gp.GP() object), this is usually log-transformed and log-baseline-subtracted
                + GP_Output (output of gp.GP().predict()), hence also log-trasnformed and log-baseline-subtracted
                + OD_Growth_Data (GP_Input but converted to real OD and centered at zero)
                + OD_Growth_Fit (GP_Output but converted to real OD and centered at zero)      
        '''

        gp_time = self.x.ravel()
        gp_input = self.y.ravel()
        gp_output = self.y0.ravel()
        gp_derivative = self.y1.ravel()

        od_growth_data = self.linear_input.ravel()
        od_growth_fit = self.linear_output.ravel()
        od_fit = self.linear_output_raised.ravel()

        data = [gp_time,
                gp_input,
                gp_output,
                gp_derivative,
                od_growth_data,
                od_growth_fit,
                od_fit,
                ]

        labels = ['Time',
                  'GP_Input',
                  'GP_Output',
                  'GP_Derivative',
                  'OD_Growth_Data',
                  'OD_Growth_Fit',
                  'OD_Fit',]

        data = pd.DataFrame(data,index=labels).T

        if self.name is not None:
            sample_id = pd.DataFrame([self.name]*data.shape[0],columns=['Sample_ID'])
            data = sample_id.join(data)

        return data


    def sample(self):
        '''
        Sample the posterior distribution of the latent function and its derivative 
            n times, estimate growth parametes for each sample, then summarize with 
            mean and standard deviation. 
        '''

        n = getValue('n_posterior_samples')

        samples0 = np.random.multivariate_normal(self.y0.ravel(),self.cov0,n)
        samples1 = np.random.multivariate_normal(self.y1.ravel(),self.cov1,n)

        list_params = []

        for ii,y0,y1 in zip(range(n),samples0,samples1):

            y0_ii = y0[:,np.newaxis]
            y1_ii = y1[:,np.newaxis]

            curve_ii = GrowthCurve(x=self.x,y=self.y,y0=y0_ii,y1=y1_ii,
                                   cov0=self.cov0,cov1=self.cov1)
            list_params.append(curve_ii.params)

        df_params = pd.DataFrame(list_params)
        df_params_avg = df_params.mean(numeric_only=True)
        df_params_std = df_params.std(numeric_only=True)

        df_params_avg.index = [f'mean({ii})' for ii in df_params_avg.index]
        df_params_std.index = [f'std({ii})' for ii in df_params_std.index]

        self.posterior = pd.concat([df_params_avg,df_params_std]).to_dict()

        return self


    def plot(self,ax_arg=None):

        if not ax_arg:
            fig,ax = plt.subplots(2,1,figsize=[6,8],sharex=True)
        else:
            ax = ax_arg

        t = self.x.ravel()
        y =  self.y.ravel()
        y0 = self.y0.rave()
        y1 = self.y1.ravel()

        xmin = 0
        xmax = int(np.ceil(t[-1]))

        ax[0].plot(t,y,lw=5,color=(0,0,0,0.65))
        ax[0].plot(t,y0,lw=5,color=(1,0,0,0.65))
        ax[1].plot(t,y1,lw=5,color=(0,0,0,0.65))

        [ii.set(fontsize=20) for ii in ax[0].get_xticklabels()+ax[0].get_yticklabels()]
        [ii.set(fontsize=20) for ii in ax[1].get_xticklabels()+ax[1].get_yticklabels()]

        ylabel = getValue('hypo_plot_y_label')
        ax[1].set_xlabel('Time',fontsize=20)
        ax[0].set_ylabel(ylabel ,fontsize=20)
        ax[1].set_ylabel(f'd/dt {ylabel}',fontsize=20)

        ax[0].set_xlim([xmin,xmax])

        if not ax_arg:
            return fig,ax
        else:
            return ax




#!/usr/bin/env python

'''
AMiGA wrapper: plotting heatmaps.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (9 functions)

# read
# clusterMap
# generate_missing_color
# getColorLegend
# group
# main
# pivot
# reduceDf
# saveDf

import os
import sys
import random
import operator
import numpy as np # type: ignore
import pandas as pd # type: ignore
import matplotlib.pyplot as plt # type: ignore
import matplotlib.patches as mpatches # type: ignore
import seaborn as sns # type: ignore

from matplotlib import rcParams # type: ignore
from matplotlib.colors import to_rgb # type: ignore

from .comm import smartPrint, tidyMessage
from .interface import checkParameterCommand
from .org import assembleFullName
from .org import isFileOrFolder
from .read import findPlateReaderFiles
from .utils import flattenList, subsetDf


rcParams.update({'font.size':20})
sns.set_style('whitegrid')


def main(args):

	verbose = args.verbose
	#directory = assemblePath(args.input,'summary')
	directory = args.input

	msg = 'AMiGA is peeking inside the summary directory'

	smartPrint('',verbose)
	smartPrint(tidyMessage(msg),verbose)
	# smartPrint(checkDirectoryNotEmpty(directory,'Summary')[1],verbose)

	criteria = checkParameterCommand(args.subset,sep=',')
	
	directory,filename = isFileOrFolder(directory,up=1)

	if filename:
		ls_files = [f'{directory}{os.sep}{filename}']
	else:
		ls_files = findPlateReaderFiles(directory,'.txt')

	full_df = read(ls_files)
	sub_df = subsetDf(full_df,criteria)
	sub_df = group(sub_df,args)
	sub_df = pivot(sub_df,args,args.value)
	sub_df = reduceDf(sub_df,args)
	clusterMap(sub_df,full_df,args,directory)
	saveDf(full_df,sub_df,args,directory)


def saveDf(full_df,sub_df,args,directory):

	if not args.save_filtered_table:
		return None

	sub_df = subsetDf(full_df,{args.y_variable:list(sub_df.index.values),
		                       args.x_variable:list(sub_df.keys().values)})

	fpath = assembleFullName(directory,'',args.output,'filtered','.txt')
	sub_df.to_csv(fpath,sep='\t',header=True,index=True)

def read(ls_fpaths):

	df_list = []

	for fpath in ls_fpaths:
		foo = pd.read_csv(fpath,sep='\t',header=0,index_col=0,low_memory=False)
		df_list.append(foo)
	df = pd.concat(df_list,sort=False)

	return df


def reduceDf(df,args):

	if args.filter is None:
		return df

	if 'OR' in args.filter:
		cmds = args.filter.split('OR')
	else:
		cmds = [args.filter]

	ls_rows = []
	ls_cols = []

	for cc in cmds:
		axis,outer,op,thresh = cc.strip().split(' ')

		# define logical operator
		delimiters = {'>=':operator.ge,
		              '<=':operator.le,
		              '>':operator.gt,
		              '<':operator.lt,
		              '=':operator.eq}
		op = delimiters[op]  

		# define outer operation
		outer_dict = {'any':any,'all':all,'':all}
		outer = outer_dict[outer]

		# define axis for outer operation
		axis_dict={'row':int(1),'col':int(0)}
		axis = axis_dict[axis]

		if axis==1:
			foo = df.loc[df.apply(lambda x: outer(op(x,float(thresh))),axis=axis),:]
			ls_rows.append(list(foo.index.values))
			ls_cols.append(list(foo.columns))
		elif axis==0:
			foo = df.loc[:,df.apply(lambda x: outer(op(x,float(thresh))),axis=axis)]
			ls_rows.append(list(foo.index.values))
			ls_cols.append(list(foo.columns))

		#ls_df.append(foo)
		#ls_df.append(df[df.apply(lambda x: outer(op(x,float(thresh))),axis=axis)])

	ls_rows = sorted(list(set(flattenList(ls_rows))))
	ls_cols = sorted(list(set(flattenList(ls_cols))))

	df = df.loc[ls_rows,ls_cols]

	#df = pd.concat(ls_df,axis=0,sort=False)

	return df


def group(df,args):

	opr = args.operation

	if opr is None: #opr = 'mean'
		return df

	df = df.groupby([args.x_variable,args.y_variable])

	if opr == 'mean':
		df = df.mean(numeric_only=True)
	elif opr == 'median':
		df = df.median(numeric_only=True)

	df = df.reset_index()

	return df

def generate_missing_color(colors):
	'''Credit to Andreas Dewes

	https://gist.github.com/adewes/5884820
	'''	

	def get_random_color(pastel_factor = 0.5):
		return [(x+pastel_factor)/(1.0+pastel_factor) for x in [random.uniform(0,1.0) for i in [1,2,3]]]
		
	def color_distance(c1,c2):
		return sum([abs(x[0]-x[1]) for x in zip(c1,c2)])
	
	def generate_new_color(existing_colors,pastel_factor = 0.5):
		max_distance = None
		best_color = None
		for i in range(0,100):
			color = get_random_color(pastel_factor = pastel_factor)
			if not existing_colors:
				return color
			best_distance = min([color_distance(color,c) for c in existing_colors])
			if not max_distance or best_distance > max_distance:
				max_distance = best_distance
				best_color = color
		return best_color

	colors = [to_rgb(ii) for ii in colors]
	best_color = tuple(generate_new_color(colors))

	return best_color


def pivot(df,args,metric=None):

	if metric is None:

		return df

	else: 

		df = pd.pivot(data=df,columns=args.x_variable,index=args.y_variable,values=metric)

		rows_todrop = np.where(df.isna().any(axis=1))[0]
		rows_todrop = df.index.values[rows_todrop]

		cols_todrop = np.where(df.isna().any(axis=0))[0]
		cols_todrop = df.keys().values[cols_todrop]

		if len(rows_todrop) > 0 or len(cols_todrop): 
			msg = 'User Warning: The heatmap data is missing values. '
			msg += 'Pleae check the data for the following:\n\n'
			msg += 'Columns:\t'
			msg += ', '.join(cols_todrop) + '\n'
			msg += '\n'
			msg += 'Rows:\t'
			msg += ', '.join(rows_todrop) + '\n'
			msg += '\nThese variables will be dropped and not plotted unless if you requested that '
			msg += 'they be kept with --keep-rows-missing-data or --keep-columns-missing-data.\n\n'
			smartPrint(msg,args.verbose)

		if not args.keep_rows_missing_data:
			df = df.drop(labels=rows_todrop,axis=0)

		if not args.keep_columns_missing_data:
			df = df.drop(labels=cols_todrop,axis=1)

	return df



def get_color_legend(df,full_df,args,directory,axis='y'):

    # e.g. color by ribotype
    colorby = eval('args.color_'+axis+'_by')
    if colorby is None:
        return None

    # variable on axis
    variable = eval('args.'+axis+'_variable')

    # e.g. file with two columns: ribotype and color
    colorfile =eval('args.color_file_'+axis)

    # dictionary arguments passed by user
    colorscheme = checkParameterCommand(eval('args.color_scheme_'+axis))
    
    # can't pass both file and command-line argument
    if colorfile is not None and colorscheme is not None:
        msg = 'WARNING: User must pass eithe color file or color scheme '
        msg += ' for the {}-axis, not both.' .format(axis)
        sys.exit(msg.format(axis))

    if colorfile is not None:
        colors_df = pd.read_csv(colorfile,sep='\t',header=0,index_col=0)
    if colorscheme is not None:
        colors_df = pd.DataFrame(colorscheme,index=['Color']).T

    # create list of colors based on meta-data
    if args.missing_color is None:
        missing_color = generate_missing_color(list(colors_df.Color.values))
    else:
        missing_color = args.missing_color

    if colorby == variable:
        foo = full_df.loc[:,[variable]]
    else:
        foo = full_df.loc[:,[colorby,variable]]

    foo = foo.drop_duplicates().set_index(variable).astype(str)
    foo = foo.join(colors_df,on=colorby)
    foo.Color = [missing_color if str(ii)=='nan' else ii for ii in list(foo.Color.values)]
    if axis == 'x':
        colors = foo.loc[df.columns,'Color'].values
    if axis == 'y':
        colors = foo.loc[df.index.values,'Color'].values

    # create legend patches
    colors_df.loc['~other~','Color'] = missing_color
    colors_df = colors_df[colors_df.Color.isin(colors)]
    colors_df = colors_df.to_dict()['Color']
    patches = [mpatches.Patch(color=color,label=label) for label,color in colors_df.items()]

    # plot legend
    fig,ax = plt.subplots(figsize=[4,4])
    ax.axis(False)
    lgd = ax.legend(handles=patches,loc='center')

    # save legend
    fpath = assembleFullName(directory,'',args.output,axis+'_legend','.pdf')
    plt.savefig(fpath,bbox_extra_artists=(lgd,), bbox_inches='tight')
    plt.close()

    return colors

def sort_heatmap(df,full_df,args,kwargs):

	if (df.shape[1] == 1) and (args.sort_y_by is None):

		kwargs['row_cluster'] = False
		df = df.sort_values(df.keys()[0],ascending=False)

	elif (args.sort_y_by == args.y_variable) and (not kwargs['row_cluster']):

		df = df.sort_index()
	
	elif (args.sort_y_by is not None) and (not kwargs['row_cluster']):

		sub_full_df = full_df.loc[:,[args.y_variable,args.sort_y_by]]
		sub_full_df = sub_full_df.drop_duplicates()
		sub_full_df = sub_full_df.set_index(args.y_variable)
		sub_full_df = df.join(sub_full_df,how='left')
		sub_full_df = sub_full_df.sort_values(args.sort_y_by)
		df =  sub_full_df.drop(labels=[args.sort_y_by],axis=1)		

	if (df.shape[0] == 0) and (args.sort_x_by is None):
		kwargs['col_cluster'] = False
		df = df.T.sort_values(df.index[0],ascending=False).T

	elif (args.sort_x_by == args.x_variable) and (not kwargs['col_cluster']):
		df = df.T.sort_index().T
	
	elif (args.sort_x_by is not None) and (not kwargs['col_cluster']):

		sub_full_df = full_df.loc[:,[args.x_variable,args.sort_x_by]]
		sub_full_df = sub_full_df.drop_duplicates()
		sub_full_df = sub_full_df.set_index(args.x_variable)
		sub_full_df = df.T.join(sub_full_df,how='left')
		sub_full_df = sub_full_df.sort_values(args.sort_x_by)
		df =  sub_full_df.drop(labels=[args.sort_x_by],axis=1).T

	return df,kwargs


def dyanmically_size_font(ax_heatmap,ax_cax,args,df):

	def get_bbox(item,ax=None,fig=None):
		r = fig.canvas.get_renderer()
		bb = item.get_window_extent(renderer=r)
		bb = bb.transformed(ax.transAxes.inverted())
		return bb

	def get_bbox_width_height(item,ax=None,fig=None):
		bb = get_bbox(item,ax=ax,fig=fig)
		return (bb.width, bb.height)

	def set_fontsize(which='x',ax=None,fig=None,fontscale=0.5,nlabels=None):
		# initialze, what is the max bbox height
		fontsize = 0	
		dim = [0 if which=='x' else 1][0]
		if nlabels is None:
			nlabels = eval(f'len(ax.get_{which}ticks())')
		eval(f'[ii.set(fontsize={fontsize}) for ii in ax.get_{which}ticklabels()]')
		max_bb_height = 1. / float(nlabels)

		# what is the current bbox height
		sample = eval(f'ax.get_{which}ticklabels()[0]')
		bb_height = get_bbox_width_height(sample,ax,fig)[dim]

		# incrementally increase bbox height until you reach target
		while bb_height < fontscale* max_bb_height:
			fontsize +=1
			eval(f'[ii.set(fontsize={fontsize}) for ii in ax.get_{which}ticklabels()]')
			bb_height = get_bbox_width_height(sample,ax,fig)[dim]

		return fontsize

	xtls = float(args.x_tick_labels_scale)
	ytls = float(args.y_tick_labels_scale)
	ctls = float(args.color_bar_labels_scale)

	if xtls != 1:
		x_fontsize = set_fontsize(which='x',ax=ax_heatmap,fig=plt.gcf(),fontscale= xtls,nlabels=df.shape[1])
		[ii.set(fontsize=x_fontsize) for ii in ax_heatmap.get_xticklabels()]

	if ytls != 1:
		y_fontsize = set_fontsize(which='y',ax=ax_heatmap,fig=plt.gcf(),fontscale= ytls,nlabels=df.shape[0])
		[ii.set(fontsize=y_fontsize) for ii in ax_heatmap.get_yticklabels()]
	
	t_fontsize = set_fontsize(which='y',ax=ax_cax,fig=plt.gcf(),fontscale= ctls)
	[ii.set(fontsize=t_fontsize) for ii in ax_cax.get_yticklabels()]

	# adjust tick labels
	if (int(args.x_rotation) % 90)==0:
		ha = 'center'
	else:
		ha = 'right'

	[ii.set(rotation=args.x_rotation,ha=ha) for ii in ax_heatmap.get_xticklabels()]



def clusterMap(df,full_df,args,directory):

	def dekwarg(ii):
		'''Defines how to dfine types of user-defined arguments'''
		key = ii.split(':')[0]
		value = ii.split(':')[1]
		adj_value = value.replace('-','')
		adj_value = adj_value.replace('.','')
		if adj_value.isdigit():
			value = float(value)
		elif value in ['True','False']:
			value = eval(value)
		return key,value

	# define figure size (inches)
	ny,nx = df.shape

	if args.width_height is None:
		figsize=[nx*2+6,ny*0.5+3]
	else:
		w,h = args.width_height
		figsize=[float(w),float(h)]

    # package argumtns into a dictionary
	kwargs = {'row_cluster':False,'col_cluster':False,'figsize':figsize}
	if args.kwargs:
		h_kwargs = args.kwargs.split(';')
		h_kwargs = [dekwarg(ii) for ii in h_kwargs]
		h_kwargs = {k:v for k,v in h_kwargs}
		kwargs.update(h_kwargs)

	if kwargs['row_cluster'] or args.cluster_y:
		kwargs['row_cluster'] = True
	if kwargs['col_cluster'] or args.cluster_x:
		kwargs['col_cluster'] = True

	# sort heatmap, if requested
	df,kwargs = sort_heatmap(df,full_df,args,kwargs)

	# get colors for side bars
	row_colors = get_color_legend(df,full_df,args,directory,axis='y')
	col_colors = get_color_legend(df,full_df,args,directory,axis='x')

	# big finale
	c = sns.clustermap(df,**kwargs,dendrogram_ratio=float(args.colorbar_ratio),
		row_colors=row_colors,col_colors=col_colors,
		colors_ratio=(args.color_x_ratio,args.color_y_ratio),
		yticklabels=True,xticklabels=True,annot=False)

	# adjust title
	title = [args.value if args.title is None else args.title][0]
	if args.color_x_by is not None:
		pad = 40
	else:
		pad = 15
	c.ax_heatmap.set_title(title,pad=pad)

	# adjust labels
	kwargs = {'xlabel':'','ylabel':''}
	c.ax_heatmap.set(**kwargs)
	c.ax_row_dendrogram.set_visible(False)
	c.ax_col_dendrogram.set_visible(False)

	# adjust color bar position and dimensions
	dendro_box = c.ax_row_dendrogram.get_position()
	dendro_box.x0 = (dendro_box.x0 + 2 * dendro_box.x1) /3
	dendro_box.x0 = dendro_box.x0 - 0.01
	dendro_box.x1 = dendro_box.x1 - 0.01
	c.cax.set_position(dendro_box)
	c.cax.yaxis.set_ticks_position("left")

	# adjust size and rotation of tick labels
	dyanmically_size_font(c.ax_heatmap,c.cax,args,df)

	# check for proper rendering of tick labels
	msg = 'WARNING: figure size is too small and /or fontsize is too large '
	msg += 'enought to display all {}-axis labels. Pleas increase the {} '
	msg += 'argument and/or decrease the fontsize to insure that all labels '
	msg += 'are propely printed'

	yticklabels = c.ax_heatmap.get_yticklabels()
	xticklabels = c.ax_heatmap.get_xticklabels()

	if df.shape[0] != len(yticklabels):
		print(msg.format('y','height'))
	if df.shape[1] != len(xticklabels):
		print(msg.format('x','width'))

	# highlight feature
	highlight_labels = checkParameterCommand(args.highlight_labels)

	if highlight_labels is not None:
		for key,values in highlight_labels.items():
			if key == 'x':
				tlabels = xticklabels
			elif key == 'y':
				tlabels = yticklabels
			for label in values:
				matches = np.where([ii.get_text()==label for ii in tlabels])[0]
				for match in matches:
					tlabels[match].set(color='red',fontweight='bold')
  

	fpath = assembleFullName(directory,'',args.output,'','.pdf')
	plt.savefig(fpath,bbox_inches='tight')


#!/usr/bin/env python

'''
AMiGA configuration file for setting and modifying default parameters used by AMiGA.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# NOTES
#     colors are defined in (R,G,B,A) format where A is alpha and all values range
#     	 from 0.0 to 1.0 (i.e. map 0 to 255) but you can define colors in text or hex
#        string format, see https://het.as.utexas.edu/HET/Software/Matplotlib/api/colors_api.html 

config = {}

###	--------------------- ###
### DATA INPUT PARAMETERS ###
###	--------------------- ###

# acceptable values are 'seconds', 'minutes', or 'hours'
config['time_input_unit'] = 'seconds'
config['time_output_unit'] = 'hours'

# default time interval between OD measurments is set to 600 seconds
config['interval'] = 600  # units ared based on 'time_input_unit' above  

# whether to estimate OD at the first time point using a polynomial regression fit across replicates
config['PolyFit'] = False

# How to handle nonpositive (i.e. zero or negative) values. See AMiGA docs for more details
config['handling_nonpositives'] = 'Delta' # 'Delta' or 'LOD'

config['limit_of_detection'] = 0.010 # must be numeric and positive (i.e. not zero)
config['force_limit_of_detection'] = False # or True. Only applies if: config['handling_nonpositives'] = 'Delta'
config['number_of_deltas'] = 5 # must be integer greater than 1, can be really large number (e.g. 1000) to comply with curves of various lengths
config['choice_of_deltas'] = 'median' # or min or max or mean

# Whether to drop blank or control wells?
config['drop_blank_wells'] = False
config['drop_control_wells'] = False

###	------------- ###
### 96-Well Plots ###
###	------------- ###

# parameters related to plotting and fold change
config['fcg'] = 1.5  # fold-change threshold for growth
config['fcd'] = 0.5  # fold-change threshold for death

config['fcg_line_color'] = (0.0,0.0,1.0,1.0)
config['fcg_face_color'] = (0.0,0.0,1.0,0.15)

config['fcd_line_color'] = (1.0,0.0,0.0,1.0)
config['fcd_face_color'] = (1.0,0.0,0.0,0.15)

config['fcn_line_color'] = (0.0,0.0,0.0,1.0)  # fc-neutral: i.e. fold-change is within thresholds defined above
config['fcn_face_color'] = (0.0,0.0,0.0,0.15)  # fc-neutral: i.e. fold-change is within thresholds defined above

config['gp_line_fit'] = 'yellow'

# parameters related to annotating grid plots with OD Max and Well ID values
config['fcn_well_id_color'] = (0.65,0.165,0.16,0.8)
config['fcn_od_max_color'] = (0.0,0.0,0.0,1.0)

# parameter for labeling y-axis of grid plots
config['grid_plot_y_label'] = 'Optical Density'

# how to handle flagged wells in plots
config['plot_flag_wells'] = 'cross' # 'empty', cross', or 'keep'

### --------- ###
### Heat Maps ###
### --------- ###

config['x_tick_labels_scale'] = 1
config['y_tick_labels_scale'] = 1
config['color_bar_labels_scale'] = 0.2

###	---------------- ###
### Model Parameters ###
###	---------------- ###

# select mode for initializing hyperparamters of GP regression
#     1. "default" will set viarance and lengthscale parameters to 1
#     2. "moments-based-proper" will initialize lengthscale to range of time and
#            will optimize variance based on a grid search
#     3. "moments-based-fast" will initilize lengthscale to range of time and will
#            initilize variance to variance of measurements
config['initialization_method'] = 'default'  # "deault", "moments-based-proper" or "moments-based-fast"

# for GP regression with input-dependent noise, select a variance smoothing window: 
config['variance_smoothing_window'] = 6 # number of x-values, based on default paramters: 6 * 600 seconds = 1 hour

# for GP regression on individual curves, AMiGA can check quality of fit by comparing the 
# estiamted carrying capacity to the actual carrying capacity calculated from data
# K_actual = OD_Max - OD_Baseline or Adj_OD_Max-Adj_OD_Basline
# here, the user can define the threshold at which AMiGA will flag poor fit of a curve. the default
# threshold is 20%.
config['k_error_threshold'] = 20

###	------------------------ ###
### Hypothesis Testing Plots ###
###	------------------------ ###

# hypothesis testing plot colors (the first two are the default colors used by AMiGA)
config['hypo_colors']  = [(0.11,0.62,0.47),(0.85,0.37,0.01)]  # correspond to seagreen

config['hypo_plot_y_label'] = 'OD'

config['HypoPlotParams'] = {'overlay_actual_data':True,
							'fontsize':15,
							'tick_spacing':5,
							'legend':'outside'}

###	----------------- ###
### GROWTH PARAMETERS ###
###	----------------- ###

# adaptationtime
config['confidence_adapt_time'] = 0.95

# diaxuic shift paramters
config['diauxie_ratio_varb'] = 'K' # can be "K" (max OD) or "r" (max dOD/dt)
config['diauxie_ratio_min'] = 0.20  # minimum ratio relative to maximum growth or growth rate for each growth phase
config['diauxie_k_min'] = 0.10
 
# parameters for reporting results (not implemented yet)
config['report_parameters'] = ['auc_lin','auc_log','k_lin','k_log','death_lin','death_log',
							   'gr','dr','td','lagC','lagP',
                               't_k','t_gr','t_dr','diauxie']
#config['report_parameters'] = ['auc_lin','k_lin','lagP','x_dr','diauxie']

# how many samples from the posterior funtion are used for estimating mean/std of growth parameters
config['n_posterior_samples'] = 100

###	------------------ ###
### USER COMMUNICATION ###
###	------------------ ###

config['Ignore_RuntimeWarning'] = True



#!/usr/bin/env python

'''
AMiGA library of auxiliary functions for plotting suppport.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"

# TABLE OF CONTENTS (7 functions)

# largeTickLabels
# savePlotWithLegends
# setAxesLabels
# dynamicWindowAdjustment
# addRealPlotLine
# addMVNPlotLines
# plotDeltaOD

import numpy as np # type: ignore
import matplotlib.pyplot as plt # type: ignore

from scipy.stats import norm # type: ignore

from .utils import subsetDf, getValue, getTimeUnits


def largeTickLabels(ax,fontsize=20):

    [ii.set(fontsize=fontsize) for ii in ax.get_xticklabels()+ax.get_yticklabels()]

    return ax


def savePlotWithLegends(ax,fig_path,loc,fontsize=20):
    '''
    Given an ax (matplotlib.axes._subplots.AxesSubplot) and path (str), save the figure
        with legend outside on the center left and keep a tight box around contents. 
    '''

    handles, labels = ax.get_legend_handles_labels()
    if loc == 'outside':
        lgd = ax.legend(handles,labels,loc='center left',bbox_to_anchor=(1,0.5),fontsize=fontsize)
    elif loc == 'inside':
        lgd = ax.legend(handles,labels,loc='lower left',fontsize=fontsize)
    plt.savefig(fig_path,bbox_extra_artists=(lgd,), bbox_inches='tight')


def setAxesLabels(ax,subtract_control,plot_params,logged=True,fontsize=20):
    ''''
    Given an axis and analysis parameters, determine appropriate labels 
        for axes and adjus them accordingly. 

    Args:
        ax (matplotlib.axes._subplots.AxesSubplot) 
        subtract_control (boolean)
        plot_params (dictionary)
        fontsize (float)

    Returns:
        ax (matplotlib.axes._subplots.AxesSubplot) 
    '''
    import matplotlib as mpl
    mpl.rcParams["mathtext.default"] = 'regular'
    mpl.rcParams["font.family"] = 'sans-serif'
    mpl.rcParams["font.sans-serif"] = 'Arial'
    # mpl.rcParams["text.usetex"] = True

    #if plot_params['plot_linear_od']:
    #    base = getValue('hypo_plot_y_label')
    #    base = r'$\frac{{{}}}{{{}}}$'.format(base+'(t)',base+'(0)')
    #else:
    if logged:
        base = 'ln {}'.format(getValue('hypo_plot_y_label'))
    else:
        base = getValue('hypo_plot_y_label')

    # plot aesthetics
    if subtract_control:
        ylabel = f'Normalized {base}'
    else:
        ylabel = base

    ax.set_xlabel('Time ({})'.format(getTimeUnits('output')),fontsize=plot_params['fontsize'])
    ax.set_ylabel(ylabel,fontsize=plot_params['fontsize'])

    return ax


def dynamicWindowAdjustment(ax):
    '''
    Adjusts the y-axis limits for figure assembled in plotHypothesis. For the to_permute
        panels, adjust all axes to accomodate most minimal and most maximal point across
        all sub-plots in the row. For the bottow panel, adjust all axes to accomodate the
        maximum absolute point while also maintaining symmetry so that the y-axis is 
        centered at zero.

    Args:
        ax (matplotlib.axes._subplots.AxesSubplot)

    Returns:
        ax (matplotlib.axes._subplots.AxesSubplot)
    '''

    def adjustWindow(ax,ymin,ymax,symmetric=False):
        '''
        Given an axis, and y-limits (ymin,ymax), and symmetry choice, adjust y-axis.
        '''
        if symmetric:
            ax.set_ylim([-1*np.ceil(ymin),np.ceil(ymax)])
        else:
            ax.set_ylim([np.floor(ymin),1*np.ceil(ymax)])
        return ax

    if isinstance(ax,np.ndarray):

        if len(ax.shape) > 1: ## conditioning

            # track min and max of plot axis limits
            y_min,y_max = 0,0
            for ax_ii in ax[0,:]:
                y_min = np.min([y_min,ax_ii.get_ylim()[0]])
                y_max = np.max([y_max,ax_ii.get_ylim()[1]])

            # adjust all axes to min and max
            for ax_ii in ax[0,:]:
                ax_ii = adjustWindow(ax_ii,y_min,y_max,symmetric=False)

            # track max of plot axis absolute limits
            y_max = 0
            for ax_ii in ax[1,:]:
                y_max = np.max([y_max,np.abs(ax_ii.get_ylim()[0]),np.abs(ax_ii.get_ylim()[1])])

            # adjust all axes to min and max
            for ax_ii in ax[1,:]:
                ax_ii = adjustWindow(ax_ii,y_max,y_max,symmetric=True)

        elif len(ax.shape) == 1: ## no conditioning
        
            # round window limits
            y_min,y_max = ax[0].get_ylim()
            ax[0] = adjustWindow(ax[0],y_min,y_max)

            # symmetrical and rounded window limits
            y_max = np.max([np.abs(ax[1].get_ylim()[0]),np.abs(ax[1].get_ylim()[1])])
            ax[1] = adjustWindow(ax[1],y_max,y_max,symmetric=True)

    else:
        # round window limits
        y_min,y_max = ax.get_ylim()
        ax = adjustWindow(ax,y_min,y_max)

    return ax


def addMVNPlotLine(ax,x,criteria,label,z_value,color,plot_params,noise=False):

    '''
    Given data (x) and criteria, find relevant sample IDs and plot them on axis.

    Args:
        ax (matplotlib.axes._subplots.AxesSubplot)
        x (pandas.DataFrame): must include columns for Time, mu, Sigma
        criteria (dictionary): keys must be column headers in x, values must be values in x.
        label (str): used for legend label of plotted line.
        z_value (float): z-value for computing confidence interval
        color (str or (R,G,B,A)) where R,G,B,A are floats [0,1]
        plot_params (dictionary)
        noise (boolean): whetehr to plot 95-pct credibel intervals including sample uncertainty

    Returns:
        ax (matplotlib.axes._subplots.AxesSubplot)    
    '''
    scaler = norm.ppf(z_value) # define confidence interval scaler for MVN predictions
    x = subsetDf(x,criteria) # grab value-specific model predictions

    if noise:
        Sigma = x.Sigma + x.Noise
    else:
        Sigma = x.Sigma

    # compute credible interval
    xtime = x.Time
    y_avg = x.mu
    y_low = x.mu-scaler*np.sqrt(Sigma)
    y_upp = x.mu+scaler*np.sqrt(Sigma)

    # convert from log2 to linear OD
    # if plot_params['plot_linear_od']:
    #     y_avg = np.exp(y_avg)
    #     y_low = np.exp(y_low)
    #     y_upp = np.exp(y_upp)

    ax.plot(xtime,y_avg,color=color,label=label,alpha=0.9,lw=3.0,zorder=10)
    ax.fill_between(x=xtime,y1=y_low,y2=y_upp,color=color,alpha=0.10,zorder=5)
    ax = largeTickLabels(ax,fontsize=plot_params['fontsize']) 

    #if plot_params['plot_linear_od']:
    #    ax.axhline(y=1,xmin=0,xmax=xtime.max(),lw=3.0,color=(0,0,0,1))
    #else:
    ax.axhline(y=0,xmin=0,xmax=xtime.max(numeric_only=True),lw=3.0,color=(0,0,0,1))

    return ax


def addRealPlotLine(ax,plate,criteria,color,plot_params):
    '''
    Given data (plate) and criteria, find relevant sample IDs and plot them on axis.

    Args:
        ax (matplotlib.axes._subplots.AxesSubplot)
        plate (GrowthPlate object)
        criteria (dictionary): keys must be column headers in plate.key, values must be 
            values in plate.key.
        color (str or (R,G,B,A)) where R,G,B,A are floats [0,1]
        plot_params (dictionary)

    Returns:
        ax (matplotlib.axes._subplots.AxesSubplot)    
    '''

    if plot_params['overlay_actual_data']:

        samples = list(subsetDf(plate.key,criteria).index)

        if len(samples)==0:
            return ax 

        time = plate.time.copy()
        data = plate.data.copy()

        #if plot_params['plot_linear_od']:
        #    data = data.apply(np.exp).copy()

        wide_df = time.join(data)

        wide_df = wide_df.reindex(['Time']+samples,axis=1).set_index('Time')
        wide_df = wide_df.dropna(axis=1) ## necessary to get rid of controls
        ax.plot(wide_df,color=color,alpha=0.5,lw=1,zorder=1)

    return ax


def plotDeltaOD(ax,df,ylabel,xlabel,fontsize=20):
    '''
    Plots Delta OD: its mean (Avg key in df) and credible intervals (Low and Upp keys). 

    Args:
        ax (matplotlib.axes._subplots.AxesSubplot)
        df (pandas.DataFrame)
        ylabel (str)
        xlabel (str)
        fontsize (float)

    Returns: 
        ax (matplotlib.axes._subplots.AxesSubplot)    
    '''

    ax.plot(df.Time,df.Avg,lw=3.0)
    ax.fill_between(df.Time,df.Low,df.Upp,alpha=0.1)
    ax.axhline(y=0,xmin=0,xmax=df.Time.max(numeric_only=True),lw=3.0,color=(0,0,0,1))
    ax = largeTickLabels(ax,fontsize=fontsize)

    if xlabel:
        ax.set_xlabel('Time ({})'.format(getTimeUnits('output')),fontsize=fontsize)
    if ylabel:
        ax.set_ylabel(r'${\Delta}$(ln OD)',fontsize=fontsize)

    return ax 



#!/usr/bin/env python

'''
AMiGA library of functions for analyzing growth curves.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (9 functions)

# basicSummaryOnly
# runGrowthFitting
# runCombinedGrowthFitting
# handleMissingData
# prepDataForFitting
# savePlots
# prepGpData
# savePlateData
# mergeSummaryData

import pandas as pd # type: ignore
import numpy as np # type: ignore
import tempfile
import sys
import os

from .model import GrowthModel 
from .growth import GrowthPlate
from .org import assembleFullName, assemblePath
from .utils import concatFileDfs, resetNameIndex, subsetDf, flattenList
from .utils import getValue, getTimeUnits, selectFileName
from .comm import smartPrint, tidyMessage, tidyDictPrint
from .params import initDiauxieList, initParamDf, initParamList, mergeDiauxieDfs
from .params import minimizeDiauxieReport, minimizeParameterReport


def basicSummaryOnly(data,mapping,directory,args,verbose=False):
    '''
    If user only requested plotting, then for  each data file, perform a basic algebraic summary
        and plot data. Once completed, exit system. Otherwise, return None.
 
    Args:
        data (dictionary): keys are plate IDs and values are pandas.DataFrames with size t x (n+1)
            where t is the number of time-points and n is number of wells (i.e. samples),
            the additional 1 is due to the explicit 'Time' column, index is uninformative.
        mapping (dictionary): keys are plate IDs and values are pandas.DataFrames with size n x (p)
            where is the number of wells (or samples) in plate, and p are the number of variables or
            parameters described in dataframe.
        directory (dictionary): keys are folder names, values are their paths
        args (argparse.Namespace class): attribute names are arguments and their value are user/default choices
        verbose (boolean)

    Returns:
        None: if only_plot_plate argument is False. 
    '''

    list_keys = []

    for pid,data_df in data.items():

        # define paths where summary and plot will be saved
        key_file_path = assemblePath(directory['summary'],pid,'.txt')
        key_fig_path = assemblePath(directory['figures'],pid,'.pdf')

        # grab plate-specific samples
        #   index should be well IDs but a      column Well should also exist
        #   in main.py, annotateMappings() is called which ensures the above is the case
        mapping_df = mapping[pid]
        mapping_df = resetNameIndex(mapping_df,'Well',False)

        # grab plate-specific data
        wells = list(mapping_df.Well.values)
        data_df = data_df.loc[:,['Time']+wells]

        # update plate-specific data with unique Sample Identifiers 
        sample_ids = list(mapping_df.index.values)
        data_df.columns = ['Time'] + sample_ids

        # create GrowthPlate object, perform basic summary
        plate = GrowthPlate(data=data_df,key=mapping_df)
        plate.convertTimeUnits(input=getTimeUnits('input'),output=getTimeUnits('output'))
        plate.computeBasicSummary()
        plate.computeFoldChange(subtract_baseline=True)
        plate.subtractControl(to_do=args.subtract_blanks,drop=False,blank=True)
        plate.subtractControl(to_do=args.subtract_control,drop=False,blank=False)

        # plot and save as PDF, also save key as TXT
        if not args.dont_plot:
            plate.plot(key_fig_path)

        if args.merge_summary:
            list_keys.append(plate.key) 
        else:
            plate.key.to_csv(key_file_path,sep='\t',header=True,index=False)

        smartPrint(pid,verbose=verbose)

    if args.merge_summary:
        filename = selectFileName(args.output)
        summary_path = assembleFullName(directory['summary'],'summary',filename,'basic','.txt')
        summary_df = pd.concat(list_keys,sort=False)
        summary_df.to_csv(summary_path,sep='\t',header=True,index=False)

    smartPrint('\nSee {} for summary text file(s).'.format(directory['summary']),verbose)
    smartPrint('See {} for figure PDF(s).\n'.format(directory['figures']),verbose)

    msg = 'AMiGA completed your request and '
    msg += 'wishes you good luck with the analysis!'
    print(tidyMessage(msg))

    sys.exit()


def runGrowthFitting(data,mapping,directory,args,verbose=False):
    '''
    Uses Gaussian Processes to fit growth curves and infer paramters of growth kinetics.  

    Args:
        data (pandas.DataFrame): number of time points (t) x number of variables plus-one (p+1)
            plus-one because Time is not an index but rather a column.
        mapping (pandas.DataFrame): number of wells/samples (n) x number of variables (p)
        directory (dictionary): keys are folder names, values are their paths
        args (argparse.Namespace class): attribute names are arguments and their value are user/default choices
        verbose (boolean)

    Action:
        saves summary text file(s) in summary folder in the parent directory.
        saves figures (PDFs) in figures folder in the parent directory.
        saves data text file(s) in derived folder in the parent directory.
    '''

    if args.pool_by:
        runCombinedGrowthFitting(data,mapping,directory,args,verbose=verbose)
        return None

    # only store data if user requested its writing or requested plotting
    if args.save_gp_data or args.plot or args.plot_derivative:
        store = True
    else:
        store = False

    #if args.output: args.merge_summary = True

    # if user requested merging of summary/data, store each plate's data/summary in temp directory first
    tmpdir = tempfile.mkdtemp()
    saved_umask = os.umask(0o77)  ## files can only be read/written by creator for security
    print(f'Temporary directory is {tmpdir}\n')

    # pre-process data
    plate = prepDataForFitting(data,mapping,subtract_baseline=True,drop_flagged_wells=False,
        subtract_control=args.subtract_control,subtract_blanks=args.subtract_blanks,log_transform=args.log_transform)
 
    ls_temp_files = []
    ls_summ_files = []
    ls_diux_files = []

    # for each plate, get samples and save individual text file for plate-specific summaries
    for pid in plate.key.Plate_ID.unique():

        smartPrint(f'Fitting {pid}',verbose)

        # grab plate-specific summary
        sub_plate = plate.extractGrowthData(args_dict={'Plate_ID':pid})

        # the primary motivation of this function: run gp model 
        sub_plate.model(nthin=args.time_step_size,store=store,verbose=verbose)

        sub_plate = cleanUnlogged(sub_plate,args)

        # save plots, if requested by user
        savePlots(sub_plate,args,directory,pid)
        
        # define file paths where data will be written
        if args.merge_summary or args.output:
            temp_path = assembleFullName(tmpdir,'',pid,'gp_data','.txt')
            summ_path = assembleFullName(tmpdir,'',pid,'summary','.txt') 
            diux_path = assembleFullName(tmpdir,'',pid,'diauxie','.txt') 
        else:
            temp_path = assembleFullName(directory['derived'],'',pid,'gp_data','.txt')
            summ_path = assembleFullName(directory['summary'],'',pid,'summary','.txt')
            diux_path = assembleFullName(directory['summary'],'',pid,'diauxie','.txt')

        # save data, if requested by user
        savePlateData(args.save_gp_data,sub_plate,temp_path,summ_path,diux_path)

        # track all potentially created files
        ls_temp_files.append(temp_path)
        ls_summ_files.append(summ_path)
        ls_diux_files.append(diux_path)

    # if user did not pass file name for output, use time stamp, see selectFileName()
    filename = selectFileName(args.output)

    # if user requested merging, merge all files in temporary directory
    mergeSummaryData(args,directory,ls_temp_files,ls_summ_files,ls_diux_files,filename)

    # remove temporary directory
    os.umask(saved_umask)
    os.rmdir(tmpdir)

    return None

def cleanUnlogged(plate,args):
    
    if args.do_not_log_transform:

        to_drop = ['auc_lin','k_lin','death_lin','td','dx_auc_lin','dx_k_lin','dx_death_lin','dx_td']
        to_drop = list(set(plate.key.columns).intersection(set(to_drop)))
        plate.key = plate.key.drop(to_drop,axis=1)
        plate.key = plate.key.rename(columns={'auc_log':'auc_lin','k_log':'k_lin','death_log':'death_lin',
            'dx_auc_log':'dx_auc_lin','dx_k_log':'dx_k_lin','dx_death_log':'dx_death_lin'})

    return plate


def runCombinedGrowthFitting(data,mapping,directory,args,verbose=False):
    '''
    Uses Gaussian Processes to fit growth curves and infer paramters of growth kinetics.
        While runGrowthFitting() analyzes data one plate at a time, runCombinedGrowthFitting()
        can pool experimental replicates across different plates. The downside is that data
        summary must be merged and no 96-well plate grid figure can be produced.  

    Args:
        data (pandas.DataFrame): number of time points (t) x number of variables plus-one (p+1)
            plus-one because Time is not an index but rather a column.
        mapping (pandas.DataFrame): number of wells/samples (n) x number of variables (p)
        directory (dictionary): keys are folder names, values are their paths
        args (argparse.Namespace class): attribute names are arguments and their value are user/default choices
        verbose (boolean)

    Action:
        saves summary text file(s) in summary folder in the parent directory.
        saves figures (PDFs) in figures folder in the parent directory.
        saves data text file(s) in derived folder in the parent directory.
    '''   

    # if user did not pass file name for output, use time stamp, see selectFileName()
    filename = selectFileName(args.output)

    # pre-process data
    plate = prepDataForFitting(data,mapping,subtract_baseline=False,drop_flagged_wells=True,
        subtract_control=args.subtract_control,subtract_blanks=args.subtract_blanks,log_transform=args.log_transform)

    # which meta-data variables do you use to group replicates?
    combine_keys = args.pool_by.split(',')
    missing_keys = [ii for ii in combine_keys if ii not in plate.key.columns]

    if missing_keys:
        msg = f'FATAL USER ERROR: The following keys {missing_keys} are '
        msg += 'missing from mapping files.'
        sys.exit(msg)

    # continue processing data
    plate.subtractBaseline(to_do=True,poly=getValue('PolyFit'),groupby=combine_keys) 
    plate_key = plate.key.copy()
    plate_data = plate.data.copy()
    plate_time = plate.time.copy()
    plate_cond = plate_key.loc[:,combine_keys+['Group','Control']].drop_duplicates(combine_keys).reset_index(drop=True)

    smartPrint(f'AMiGA detected {plate_cond.shape[0]} unique conditions.\n',verbose)

    data_ls, diauxie_dict = [], {}

    # get user-defined values from config.py
    posterior = args.sample_posterior
    fix_noise = args.fix_noise
    nthin = args.time_step_size

    # initialize empty dataframes for storing growth parameters
    params_latent = initParamDf(plate_cond.index,complexity=0)
    params_sample = initParamDf(plate_cond.index,complexity=1)
    mse_df = pd.DataFrame(index=plate_cond.index,columns=['MSE'])

    # for each unique condition based on user request
    for idx,condition in plate_cond.iterrows():

        # get list of sample IDs
        cond_dict = condition.drop(['Group','Control'])
        cond_dict = cond_dict.to_dict() # e.g. {'Substate':['D-Trehalose'],'PM':[1]} 
        cond_idx = subsetDf(plate_key,cond_dict).index.values  # list of index values for N samples
        smartPrint(f'Fitting\n{tidyDictPrint(cond_dict)}',verbose)

        # get data and format for GP instance
        cond_data = plate_data.loc[:,list(cond_idx)] # T x N
        cond_data = plate_time.join(cond_data) # T x N+1

        # warn user if there is missing data
        cond_data = handleMissingData(args,cond_data)

        cond_data = cond_data.melt(id_vars='Time',var_name='Sample_ID',value_name='OD')
        cond_data = cond_data.drop(['Sample_ID'],axis=1) # T*R x 2 (where R is number of replicates)
        cond_data = cond_data.dropna(axis=0)
        
        gm = GrowthModel(df=cond_data,ARD=True,heteroscedastic=fix_noise,nthin=nthin,logged=plate.mods.logged)#,

        curve = gm.run(name=idx)
        mse_df.loc[idx,'MSE'] = curve.compute_mse(pooled=True)

        # get parameter estimates using latent function
        diauxie_dict[idx] = curve.params.pop('df_dx')
        params_latent.loc[idx,:] = curve.params

        # get parameter estimates using samples fom the posterior distribution
        if posterior:
            params_sample.loc[idx,:] = curve.sample().posterior

        # passively save data, manipulation occurs below (input OD, GP fit, & GP derivative)
        if args.save_gp_data:
            time = pd.DataFrame(gm.x_new,columns=['Time'])
            mu0,var0 = np.ravel(gm.y0),np.ravel(np.diag(gm.cov0))
            mu1,var1 = np.ravel(gm.y1),np.ravel(np.diag(gm.cov1))

            if fix_noise:
                sigma_noise = np.ravel(gm.error_new)+gm.noise
            else:
                sigma_noise = np.ravel([gm.noise]*time.shape[0])

            mu_var = pd.DataFrame([mu0,var0,mu1,var1,sigma_noise],index=['mu','Sigma','mu1','Sigma1','Noise']).T
            gp_data = pd.DataFrame([list(condition.values)]*len(mu0),columns=condition.keys())
            gp_data = gp_data.join(time).join(mu_var)
            data_ls.append(gp_data)

    # summarize diauxie results
    diauxie_df = mergeDiauxieDfs(diauxie_dict)

    if posterior:
        gp_params = params_sample.join(params_latent['diauxie'])
    else:
        gp_params = params_latent

    # record results in object's key
    plate_cond = plate_cond.join(gp_params).join(mse_df)
    plate_cond.index.name = 'Sample_ID'
    plate_cond = plate_cond.reset_index(drop=False)
    plate_cond = pd.merge(plate_cond,diauxie_df,on='Sample_ID')

    params = initParamList(0) + initParamList(1)
    params = list(set(params).intersection(set(plate_cond.keys())))

    df_params = plate_cond.drop(initDiauxieList(),axis=1).drop_duplicates()
    df_diauxie = plate_cond[plate_cond.diauxie==1]
    df_diauxie = df_diauxie.drop(params,axis=1)
    df_diauxie = minimizeDiauxieReport(df_diauxie)

    # because pooling, drop linear AUC, K, and Death 
    to_remove = ['death_lin','k_lin','auc_lin']
    to_remove = flattenList([[jj for jj in df_params.keys().values if ii in jj] for ii in to_remove])
    df_params.drop(to_remove,axis=1,inplace=True)

    if args.do_not_log_transform: 
        to_remove = ['td','mean(td)','std(td)']
        to_remove = flattenList([[ii for jj in df_params.keys().values if ii == jj] for ii in to_remove])
        df_params.drop(to_remove,axis=1,inplace=True)

    to_remove = flattenList([[jj for jj in df_diauxie.keys().values if ii in jj] for ii in to_remove])
    df_diauxie.drop(to_remove,axis=1,inplace=True)

    if args.do_not_log_transform:
        columns = {'auc_log':'auc_lin','k_log':'k_lin','death_log':'death_lin',
                   'dx_auc_log':'dx_auc_lin','dx_k_log':'dx_k_lin','dx_death_log':'dx_death_lin'}
        df_params.rename(columns=columns,inplace=True)
        df_diauxie.rename(columns=columns,inplace=True)

    summ_path = assembleFullName(directory['summary'],'',filename,'summary','.txt')
    diux_path = assembleFullName(directory['summary'],'',filename,'diauxie','.txt')

    # normalize parameters, if requested
    df_params = minimizeParameterReport(df_params)

    # save results
    df_params.to_csv(summ_path,sep='\t',header=True,index=False)
    if df_diauxie.shape[0]>0:
        df_diauxie.to_csv(diux_path,sep='\t',header=True,index=False)

    # save latent functions
    if args.save_gp_data:
        file_path = assembleFullName(directory['derived'],'',filename,'gp_data','.txt')
        gp_data = pd.concat(data_ls,sort=False).reset_index(drop=True)
        gp_data.to_csv(file_path,sep='\t',header=True,index=True)

    return None


def handleMissingData(args,df):
    '''
    Warns use about missing data and drops it if requested

    Args:
        args (argparse.Namespace class): attribute names are arguments and their value are user/default choices
        df (pandas.DataFrame)

    Returns 
        df (pandas.DataFrame)
    '''

    missing = np.unique(np.where(df.iloc[:,1:].isna())[0])
    missing = df.iloc[missing,].Time.values

    if len(missing) > 0:
        
        msg = 'WARNING: AMiGA detected missing data. Values are missing for some '
        msg += 'replicates at the following time-points:\n'
        msg += ', '.join([str(ii) for ii in missing])
        msg += '\n'
        smartPrint(msg,args.verbose)

    if ~args.keep_missing_time_points:
        df = df[~df.isna().any(axis=1)]

    return df


def prepDataForFitting(data,mapping,subtract_baseline=True,subtract_control=False,subtract_blanks=False,log_transform=False,drop_flagged_wells=False):
    '''
    Packages data set into a grwoth.GrowthPlate() object and transforms data in preparation for GP fitting.

    Args:
        data (pandas.DataFrame): number of time points (t) x number of variables plus-one (p+1)
            plus-one because Time is not an index but rather a column.
        mapping (pandas.DataFrame): number of wells/samples (n) x number of variables (p)
       
    Returns:
        plate (growth.GrwothPlate() object) 
    '''

    # merge data-sets for easier analysis and perform basic summaries and manipulations
    plate = GrowthPlate(data=data,key=mapping)

    plate.convertTimeUnits(input=getTimeUnits('input'),output=getTimeUnits('output'))
    plate.computeBasicSummary()
    plate.computeFoldChange(subtract_baseline=subtract_baseline)
    plate.subtractControl(to_do=subtract_blanks,drop=getValue('drop_blank_wells'),blank=True)
    plate.subtractControl(to_do=subtract_control,drop=getValue('drop_control_wells'),blank=False)
    plate.raiseData()  # replace non-positive values, necessary prior to log-transformation
    plate.logData(to_do=log_transform)  # natural-log transform
    plate.subtractBaseline(subtract_baseline,poly=False)  # subtract first T0 (or rather divide by first T0)
    plate.dropFlaggedWells(to_do=drop_flagged_wells)

    return plate


def savePlots(plate,args,directory,filename):
    '''
    Saves the GP model fits of grwoth.GrowthPlate() object as a plot. 

    Args:
        plate (growth.GrwothPlate() object)
        args (argparse.Namespace class): attribute names are arguments and their value are user/default choices
        directory (dictionary): keys are folder names, values are their paths
        filename (str): file name

    Returns:
        None
    '''

    if args.plot:  # plot OD and its GP estimate

        fig_path = assembleFullName(directory['figures'],'',filename,'fit','.pdf')
        plate.plot(fig_path,plot_fit=True,plot_derivative=False)

    if args.plot_derivative:  # plot GP estimate of dOD/dt (i.e. derivative)

        fig_path = assembleFullName(directory['figures'],'',filename,'derivative','.pdf')
        plate.plot(fig_path,plot_fit=False,plot_derivative=True)


def prepGpData(plate):
    '''
    Packages different variants of grwoth curves for a specific plate into a single pandas.DataFrame.

    Args:
        plate (growth.GrwothPlate() object)

    Returns: 
        pandas.DataFrame (10 columns). See gp_data attribute in GrwothPlate().__init__() for more info.
    '''

    col_order = ['Plate_ID','Sample_ID','Time','OD_Data','OD_Fit']
    col_order += ['GP_Input','GP_Output','OD_Growth_Data','OD_Growth_Fit','GP_Derivative']

    pids = plate.key.drop_duplicates(subset='Sample_ID').set_index('Sample_ID')
    pids = pids.loc[:,['Plate_ID']]

    data = plate.time.join(plate.input_data)
    data = data.melt(id_vars='Time',var_name='Sample_ID',value_name='OD_Data')
    data = data.join(pids,on='Sample_ID')

    data = pd.merge(data,plate.gp_data,on=['Sample_ID','Time'])
    data = data.sort_values(['Sample_ID','Time'],ascending=[True,True])
    data = data.loc[:,col_order]

    return data


def savePlateData(store_data,plate,data_path,summ_path,diux_path):
    '''
    Saves the 'gp_data' and 'key' (attributes of 'plate' object) as tab-delimited files. 

    Args:
        store_data (boolean): whether to run part of code or not (save data from gp-model)
        plate (growth.GrwothPlate() object)
        data_path (str): file path for storing data file
        summ_path (str): file path for storing summary file
    '''

    #params = getValue('params_report')
    params = initParamList(0) + initParamList(1) + initParamList(2) 
    params = list(set(params).intersection(set(plate.key.keys())))
    diauxie = list(set(plate.key.keys()).intersection(set(initDiauxieList())))

    df_params = plate.key.drop(diauxie,axis=1).drop_duplicates().reset_index(drop=True)
    df_params = minimizeParameterReport(df_params)
    df_params = df_params.drop_duplicates()

    df_diauxie = plate.key[plate.key.diauxie==1]
    df_diauxie = df_diauxie.drop(params,axis=1)
    df_diauxie = df_diauxie.loc[:,~df_diauxie.columns.str.startswith('norm_')]
    df_diauxie = df_diauxie.reset_index(drop=True)
    df_diauxie = minimizeDiauxieReport(df_diauxie)


    df_params.drop(['Subset'],axis=1,inplace=True)
    df_params.to_csv(summ_path,sep='\t',header=True,index=False)
    if df_diauxie.shape[0]>0:
        df_diauxie.drop(['Subset'],axis=1,inplace=True)
        df_diauxie.to_csv(diux_path,sep='\t',header=True,index=False)

    if not store_data:
        return None

    prepGpData(plate).to_csv(data_path,sep='\t',header=True,index=True)


def mergeSummaryData(args,directory,ls_temp_files,ls_summ_files,ls_diux_files,filename):
    '''
    Reads files passed via list and concatenates them into a single pandas.DataFrame. This 
        is executed for summary files and gp_data files separately. 

    Args:
        args (argparse.Namespace class): attribute names are arguments and their value are user/default choices
        directory (dictionary): keys are folder names, values are their paths
        ls_temp_files (list): where each item is a file path (str)
        ls_temp_files (list): where each item is a file path (str)
        filename (str): base file name  
    '''

    if args.save_gp_data and (args.merge_summary or args.output):

        gp_data_df = concatFileDfs(ls_temp_files)

        gp_data_path = assembleFullName(directory['derived'],'',filename,'gp_data','.txt')

        gp_data_df.to_csv(gp_data_path,sep='\t',header=True,index=True)

    if args.merge_summary or args.output:

        summ_df = concatFileDfs(ls_summ_files)
        diux_df = concatFileDfs(ls_diux_files)

        summ_path = assembleFullName(directory['summary'],'',filename,'summary','.txt')
        diux_path = assembleFullName(directory['summary'],'',filename,'diauxie','.txt')

        summ_df.to_csv(summ_path,sep='\t',header=True,index=True)
        if diux_df.shape[0]>0:
            diux_df.to_csv(diux_path,sep='\t',header=True,index=True)

        # clean-up
        for f in ls_temp_files + ls_summ_files + ls_diux_files:
            if os.path.isfile(f):
                os.remove(f)




#/usr/env/bin python

'''
AMiGA variables that define the layout of Biolog PM Plates.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"

#### ACKNOWELDGMENT: ADOPTED FROM JAMES COLLINS. See NewPlateReader2018.py

Carbon1 = ["Negative Control",
         "L-Arabanose",
         "N-Acetyl-D-Glucosamine",
         "D-Saccharic Acid",
         "Succinic Acid",
         "D-Galactose",
         "L-Aspartic Acid",
         "L-Proline",
         "D-Alanine",
         "D-Trehalose",
         "D-Mannose",
         "Dulcitol",
         "D-Serine",
         "D-Sorbitol",
         "Glycerol",
         "L-Fucose",
         "D-Glucuronic Acid",
         "D-Gluconic Acid",
         "D,L-alpha-Glycerol-Phosphate",
         "D-Xylose",
         "L-Lactic Acid",
         "Formic Acid",
         "D-Mannitol",
         "L-Glutamic Acid",
         "D-Glucose-6-Phospate",
         "D-Galactonic Acid-gamma-Lactone",
         "D,L-Malic Acid",
         "D-Ribose",
         "Tween 20",
         "L-Rhamnose",
         "D-Fructose",
         "Acetic Acid",
         "alpha-D-Glucose",
         "Maltose",
         "D-Melibiose",
         "Thymidine",
         "L-Asparagine",
         "D-Aspartic Acid",
         "D-Glucosaminic Acid",
         "1,2-Propanediol",
         "Tween 40",
         "alpha-Keto-Glutaric Acid",
         "alpha-Keto-Butyric Acid",
         "alpha-Methyl-D-Galactoside",
         "alpha-D-Lactose",
         "Lactulose",
         "Sucrose",
         "Uridine",
         "L-Glutamine",
         "m-Tartaric Acid",
         "D-Glucose-1-Phosphate",
         "D-Fructose-6-Phosphate",
         "Tween 80",
         "alpha-Hydroxy Glutaric Acid-gamma-Lactone",
         "alpha-Hydroxy Butyric Acid", 
         "beta-Methyl-D-Glucoside",
         "Adonitol",
         "Maltotriose",
         "2-Deoxy Adenosine",
         "Adenosine",
         "Glycyl-L-Aspartic Acid",
         "Citric Acid",
         "m-Inositol",
         "D-Threonine",
         "Fumaric Acid",
         "Bromo Succinic Acid",
         "Propionic Acid",
         "Mucic Acid",
         "Glycolic Acid",
         "Glyoxylic Acid",
         "D-Cellobiose",
         "Inosine",
         "Glycyl-L-Glutamic Acid",
         "Tricarballylic Acid",
         "L-Serine",
         "L-Threonine",
         "L-Alanine",
         "L-Alanyl-Glycine",
         "Acetoacetic Acid",
         "N-Acetyl-beta-D-Mannosamine",
         "Mono Methyl Succinate",
         "Methyl Pyruvate",
         "D-Malic Acid",
         "L-Malic Acid",
         "Glycyl-L-Proline",
         "p-Hydroxy-Phenyl Acetic Acid",
         "m-Hydroxy-Phenyl Acetic Acid",
         "Tyramine",
         "D-Psicose",
         "L-Lyxose",
         "Glucuronamide",
         "Pyruvic Acid",
         "L-Galactonic Acid-gamma-Lactone",
         "D-Galacturonic Acid",
         "Phenylethylamine",
         "2-Aminoethanol"]

Carbon2 = ["Negative Control",
         "Chondroitin Sulfate C",
         "alpha-Cyclodextrin",
         "beta-Cyclodextrin",
         "gamma-Cyclodextrin",
         "Dextrin",
         "Gelatin",
         "Glycogen",
         "Inulin",
         "Laminarin",
         "Mannan",
         "Pectin",
         "N-Acetyl-D-Galactosamine",
         "N-Acetyl-Neuraminic-Acid",
         "beta-D-Allose",
         "Amygdalin",
         "D-Arabinose",
         "D-Arabitol",
         "L-Arabitol",
         "Arbutin",
         "2-Deoxy-D Ribose",
         "i-Erythritol",
         "D-Fucose",
         "3-0-beta-D-Galactopyranosyl-D Arabinose",
         "Gentiobiose",
         "L-Glucose",
         "Lactitol",
         "D-Melezitose",
         "Maltitol",
         "a-Methyl-D Glucoside",
         "beta-Methyl-D-Galactoside",
         "3-Methyl Glucose",
         "beta-Methyl-D-Glucuronic Acid",
         "alpha-Methyl-D-Mannoside",
         "beta-Methyl-D-Xyloside",
         "Palatinose",
         "D-Raffinose",
         "Salicin",
         "Sedoheptulosan",
         "L-Sorbose",
         "Stachyose",
         "D-Tagatose",
         "Turanose",
         "Xylitol",
         "N-Acetyl-D-Glucosaminitol",
         "gamma-Amino Butyric Acid",
         "delta-Amino Valeric Acid",
         "Butyric Acid",
         "Capric Acid",
         "Caproic Acid",
         "Citraconic Acid",
         "Citramalic Acid",
         "D-Glucosamine",
         "2-Hydroxy Benzoic Acid",
         "4-Hydroxy Benzoic Acid",
         "beta-Hydroxy Butyric Acid",
         "gamma-Hydroxy Butyric Acid",
         "a-Keto-Valeric Acid",
         "Itaconic Acid",
         "5-Keto-D-Gluconic Acid",
         "D-Lactic Acid Methyl Ester",
         "Malonic Acid",
         "Melibionic Acid",
         "Oxalic Acid",
         "Oxalomalic Acid",
         "Quinic Acid",
         "D-Ribono-1,4-Lactone",
         "Sebacic Acid",
         "Sorbic Acid",
         "Succinamic Acid",
         "D-Tartaric Acid",
         "L-Tartaric Acid",
         "Acetamide",
         "L-Alaninamide",
         "N-Acetyl-L-Glutamic Acid",
         "L-Arginine",
         "Glycine",
         "L-Histidine",
         "L-Homoserine",
         "Hydroxy-L-Proline",
         "L-Isoleucine",
         "L-Leucine",
         "L-Lysine",
         "L-Methionine",
         "L-Ornithine",
         "L-Phenylalanine",
         "L-Pyroglutamic Acid",
         "L-Valine",
         "D,L-Carnitine",
         "Sec-Butylamine",
         "D.L-Octopamine",
         "Putrescine",
         "Dihydroxy Acetone",
         "2,3-Butanediol",
         "2,3-Butanedione",
         "3-Hydroxy 2-Butanone"]

Nitrogen = ['Negative Control',
        'Ammonia',
        'Nitrite',
        'Nitrate',
        'Urea',
        'Biuret',
        'L-Alanine',
        'L-Arginine',
        'L-Asparagine',
        'L-Aspartic Acid',
        'L-Cysteine',
        'L-Glutamic Acid',
        'L-Glutamine',
        'Glycine',
        'L-Histidine',
        'L-Isoleucine',
        'L-Leucine',
        'L-Lysine',
        'L-Methionine',
        'L-Phenylalanine',
        'L-Proline',
        'L-Serine',
        'L-Threonine',
        'L-Tryptophan',
        'L-Tyrosine',
        'L-Valine',
        'D-Alanine',
        'D-Asparagine',
        'D-Aspartic Acid',
        'D-Glutamic Acid',
        'D-Lysine',
        'D-Serine',
        'D-Valine',
        'L-Citrulline',
        'L-Homoserine',
        'L-Ornithine',
        'N-Acetyl-L-Glutamic Acid',
        'N-Phthaloyl-L-Glutamic Acid',
        'L-Pyroglutamic Acid',
        'Hydroxylamine',
        'Methylamine',
        'N-Amylamine',
        'N-Butylamine',
        'Ethylamine',
        'Ethanolamine',
        'Ethylenediamine',
        'Putrescine',
        'Agmatine',
        'Histamine',
        'beta-Phenylethylamine',
        'Tyramine',
        'Acetamide',
        'Formamide',
        'Glucuronamide',
        'D,L-Lactamide',
        'D-Glucosamine',
        'D-Galactosamine',
        'D-Mannosamine',
        'N-Acetyl-D-Glucosamine',
        'N-Acetyl-D-Galactosamine',
        'N-Acetyl-D-Mannosamine',
        'Adenine',
        'Adenosine',
        'Cytidine',
        'Cytosine',
        'Guanine',
        'Guanosine',
        'Thymine',
        'Thymidine',
        'Uracil',
        'Uridine',
        'Inosine',
        'Xanthine',
        'Xanthosine',
        'Uric Acid',
        'Alloxan',
        'Allantoin',
        'Parabanic Acid',
        'D,L-alpha-Amino-N-Butyric Acid',
        'gamma-Amino-N-Butyric Acid',
        'g-Amino-N-Caproic Acid',
        'D,L-alpha-Amino Caprylic Acid',
        'delta-Amino-N-Valeric Acid',
        'alpha-Amino-N-Valeric Acid',
        'Ala-Asp',
        'Ala-Gln',
        'Ala-Glu',
        'Ala-Gly',
        'Ala-His',
        'Ala-Leu',
        'Ala-Thr',
        'Gly-Asn',
        'Gly-Gln',
        'Gly-Glu',
        'Gly-Met',
        'Met-Al']

PhosphorusAndSulfur = ["Negative Control",
        "Phosphate",
        "Pyrophosphate",
        "Trimetaphosphate",
        "Tripolyphosphate",
        "Triethyl Phosphate",
        "Hypophosphite",
        "Adenosine- 2'-monophosphate",
        "Adenosine- 3'-monophosphate",
        "Adenosine- 5'-monophosphate",
        "Adenosine- 2',3'-cyclic monophosphate",
        "Adenosine- 3',5'-cyclic monophosphate",
        "Thiophosphate",
        "Dithiophosphate",
        "D,L-alpha-Glycerol-Phosphate",
        "beta-Glycerol-Phosphate",
        "Carbamyl-Phosphate",
        "D-2-Phospho-Glyceric Acid",
        "D-3-Phospho-Glyceric Acid",
        "Guanosine-2'-monophosphate",
        "Guanosine-3'-monophosphate",
        "Guanosine-5'-monophosphate",
        "Guanosine-2',3'-cyclic monophosphate",
        "Guanosine-3',5'-cyclic monophosphate",
        "Phosphoenol Pyruvate",
        "Phospho-Glycolic Acid",
        "D-Glucose-1-Phosphate",
        "D-Glucose-6-Phosphate",
        "2-Deoxy-D-Glucose-6-Phosphate",
        "D-Glucosamine-6-Phosphate",
        "6-Phospho-Gluconic Acid",
        "Cytidine-2'-monophosphate",
        "Cytidine-3'-monophosphate",
        "Cytidine-5'-monophosphate",
        "Cytidine-2',3'-cyclic monophosphate",
        "Cytidine-3',5'-cyclic monophosphate",
        "D-Mannose-1-Phosphate",
        "D-Mannose-6-Phosphate",
        "Cysteamine-S-Phosphate",
        "Phospho-L-Arginine",
        "O-Phospho-D-Serine",
        "O-Phospho-L-Serine",
        "O-Phospho-L-Threonine",
        "Uridine-2'-monophosphate",
        "Uridine-3'-monophosphate",
        "Uridine-5'-monophosphate",
        "Uridine-2',3'-cyclic monophosphate",
        "Uridine-3',5'-cyclic monophosphate",
        "O-Phospho-D-Tyrosine",
        "O-Phospho-L-Tyrosine",
        "Phosphocreatine",
        "Phosphoryl Choline",
        "O-Phosphoryl-Ethanolamine",
        "Phosphono Acetic Acid",
        "2-Aminoethyl Phosphonic Acid",
        "Methylene Diphosphonic Acid",
        "Thymidine-3'-monophosphate",
        "Thymidine-5'-monophosphate",
        "Inositol Hexaphosphate",
        "Thymidine-3',5'-cyclic-monophosphate",
        "Negative Control",
        "Sulfate",
        "Thiosulfate",
        "Tetrathionate",
        "Thiophosphate",
        "Dithiophosphate",
        "L-Cysteine",
        "D-Cysteine",
        "L-Cysteinyl-Glycine",
        "L-Cysteic Acid",
        "Cysteamine",
        "L-Cysteine Sulfinic Acid",
        "N-Acetyl-L-Cysteine",
        "S-Methyl-L-Cysteine",
        "Cystathionine",
        "Lanthionine",
        "Glutathione",
        "D,L-Ethionine",
        "L-Methionine",
        "D-Methionine",
        "Glycyl-L-Methionine",
        "N-Acetyl-D,L-Methionine",
        "L- Methionine Sulfoxide",
        "L-Methionine Sulfone",
        "L-Djenkolic Acid",
        "Thiourea",
        "1-Thio-beta-D-Glucose",
        "D,L-Lipoamide",
        "Taurocholic Acid",
        "Taurine",
        "Hypotaurine",
        "p-Amino Benzene Sulfonic Acid",
        "Butane Sulfonic Acid",
        "2-Hydroxyethane Sulfonic Acid",
        "Methane Sulfonic Acid",
        "Tetramethylene Sulfone"]

PeptideNitrogen1 = ["Negative Control",
        "Positive Control: L-Glutamine",
        "Ala-Ala",
        "Ala-Arg",
        "Ala-Asn",
        "Ala-Glu",
        "Ala-Gly",
        "Ala-His",
        "Ala-Leu",
        "Ala-Lys",
        "Ala-Phe",
        "Ala-Pro",
        "Ala-Ser",
        "Ala-Thr",
        "Ala-Trp",
        "Ala-Tyr",
        "Arg-Ala",
        "Arg-Arg",
        "Arg-Asp",
        "Arg-Gln",
        "Arg-Glu",
        "Arg-Ile",
        "Arg-Leu",
        "Arg-Lys",
        "Arg-Met",
        "Arg-Phe",
        "Arg-Ser",
        "Arg-Trp",
        "Arg-Tyr",
        "Arg-Val",
        "Asn-Glu",
        "Asn-Val",
        "Asp-Asp",
        "Asp-Glu",
        "Asp-Leu",
        "Asp-Lys",
        "Asp-Phe",
        "Asp-Trp",
        "Asp-Val",
        "Cys-Gly",
        "Gln-Gln",
        "Gln-Gly",
        "Glu-Asp",
        "Glu-Glu",
        "Glu-Gly",
        "Glu-Ser",
        "Glu-Trp",
        "Glu-Tyr",
        "Glu-Val",
        "Gly-Ala",
        "Gly-Arg",
        "Gly-Cys",
        "Gly-Gly",
        "Gly-His",
        "Gly-Leu",
        "Gly-Lys",
        "Gly-Met",
        "Gly-Phe",
        "Gly-Pro",
        "Gly-Ser",
        "Gly-Thr",
        "Gly-Trp",
        "Gly-Tyr",
        "Gly-Val",
        "His-Asp",
        "His-Gly",
        "His-Leu",
        "His-Lys",
        "His-Met",
        "His-Pro",
        "His-Ser",
        "His-Trp",
        "His-Tyr",
        "His-Val",
        "Ile-Ala",
        "Ile-Arg",
        "Ile-Gln",
        "Ile-Gly",
        "Ile-His",
        "Ile-Ile",
        "Ile-Met",
        "Ile-Phe",
        "Ile-Pro",
        "Ile-Ser",
        "lle-Trp",
        "Ile-Tyr",
        "Ile-Val",
        "Leu-Ala",
        "Leu-Arg",
        "Leu-Asp",
        "Leu-Glu",
        "Leu-Gly",
        "Leu-Ile",
        "Leu-Leu",
        "Leu-Met",
        "Leu-Ph"]

PeptideNitrogen2 = ["Negative Control",
        "Positive Control: L-Glutamine",
        "Leu-Ser",
        "Leu-Trp",
        "Leu-Val",
        "Lys-Ala",
        "Lys-Arg",
        "Lys-Glu",
        "Lys-Ile",
        "Lys-Leu",
        "Lys-Lys",
        "Lys-Phe",
        "Lys-Pro",
        "Lys-Ser",
        "Lys-Thr",
        "Lys-Trp",
        "Lys-Tyr",
        "Lys-Val",
        "Met-Arg",
        "Met-Asp",
        "Met-Gln",
        "Met-Glu",
        "Met-Gly",
        "Met-His",
        "Met-Ile",
        "Met-Leu",
        "Met-Lys",
        "Met-Met",
        "Met-Phe",
        "Met-Pro",
        "Met-Trp",
        "Met-Val",
        "Phe-Ala",
        "Phe-Gly",
        "Phe-Ile",
        "Phe-Phe",
        "Phe-Pro",
        "Phe-Ser",
        "Phe-Trp",
        "Pro-Ala",
        "Pro-Asp",
        "Pro-Gln",
        "Pro-Gly",
        "Pro-Hyp",
        "Pro-Leu",
        "Pro-Phe",
        "Pro-Pro",
        "Pro-Tyr",
        "Ser-Ala",
        "Ser-Gly",
        "Ser-His",
        "Ser-Leu",
        "Ser-Met",
        "Ser-Phe",
        "Ser-Pro",
        "Ser-Ser",
        "Ser-Tyr",
        "Ser-Val",
        "Thr-Ala",
        "Thr-Arg",
        "Thr-Glu",
        "Thr-Gly",
        "Thr-Leu",
        "Thr-Met",
        "Thr-Pro",
        "Trp-Ala",
        "Trp-Arg",
        "Trp-Asp",
        "Trp-Glu",
        "Trp-Gly",
        "Trp-Leu",
        "Trp-Lys",
        "Trp-Phe",
        "Trp-Ser",
        "Trp-Trp",
        "Trp-Tyr",
        "Tyr-Ala",
        "Tyr-Gln",
        "Tyr-Glu",
        "Tyr-Gly",
        "Tyr-His",
        "Tyr-Leu",
        "Tyr-Lys",
        "Tyr-Phe",
        "Tyr-Trp",
        "Tyr-Tyr",
        "Val-Arg",
        "Val-Asn",
        "Val-Asp",
        "Val-Gly",
        "Val-His",
        "Val-Ile",
        "Val-Leu",
        "Val-Tyr",
        "Val-Val",
        "gamma-Glu-Gl"]

PeptideNitrogen3 = ["Negative Control",
        "Positive Control: L-Glutamine",
        "Ala-Asp",
        "Ala-Gln",
        "Ala-lle",
        "Ala-Met",
        "Ala-Val",
        "Asp-Ala",
        "Asp-Gln",
        "Asp-Gly",
        "Glu-Ala",
        "Gly-Asn",
        "Gly-Asp",
        "Gly-lle",
        "His-Ala",
        "His-Glu",
        "His-His",
        "Ile-Asn",
        "Ile-Leu",
        "Leu-Asn",
        "Leu-His",
        "Leu-Pro",
        "Leu-Tyr",
        "Lys-Asp",
        "Lys-Gly",
        "Lys-Met",
        "Met-Thr",
        "Met-Tyr",
        "Phe-Asp",
        "Phe-Glu",
        "Gln-Glu",
        "Phe-Met",
        "Phe-Tyr",
        "Phe-Val",
        "Pro-Arg",
        "Pro-Asn",
        "Pro-Glu",
        "Pro-lle",
        "Pro-Lys",
        "Pro-Ser",
        "Pro-Trp",
        "Pro-Val",
        "Ser-Asn",
        "Ser-Asp",
        "Ser-Gln",
        "Ser-Glu",
        "Thr-Asp",
        "Thr-Gln",
        "Thr-Phe",
        "Thr-Ser",
        "Trp-Val",
        "Tyr-lle",
        "Tyr-Val",
        "Val-Ala",
        "Val-Gln",
        "Val-Glu",
        "Val-Lys",
        "Val-Met",
        "Val-Phe",
        "Val-Pro",
        "Val-Ser",
        "beta-Ala-Ala",
        "beta-Ala-Gly",
        "beta-Ala-His",
        "Met-beta-Ala",
        "beta-Ala-Phe",
        "D-Ala-D-Ala",
        "D-Ala-Gly",
        "D-Ala-Leu",
        "D-Leu-D-Leu",
        "D-Leu-Gly",
        "D-Leu-Tyr",
        "gamma-Glu-Gly",
        "gamma-D-Glu-Gly",
        "Gly-D-Ala",
        "Gly-D-Asp",
        "Gly-D-Ser",
        "Gly-D-Thr",
        "Gly-D-Val",
        "Leu-beta-Ala",
        "Leu-D-Leu",
        "Phe-beta-Ala",
        "Ala-Ala-Ala",
        "D-Ala-Gly-Gly",
        "Gly-Gly-Ala",
        "Gly-Gly-D-Leu",
        "Gly-Gly-Gly",
        "Gly-Gly-lle",
        "Gly-Gly-Leu",
        "Gly-Gly-Phe",
        "Val-Tyr-Val",
        "Gly-Phe-Phe",
        "Leu-Gly-Gly",
        "Leu-Leu-Leu",
        "Phe-Gly-Gly",
        "Tyr-Gly-Gl"]


PM_LIST = [Carbon1,
          Carbon2,
          Nitrogen,
          PhosphorusAndSulfur,
          PeptideNitrogen1,
          PeptideNitrogen2,
          PeptideNitrogen3]




#!/usr/bin/env python

'''
AMiGA wrapper: computing confidene interval for summary or data files.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (3 functions)

# main
# get_parameter_confidence
# get_curve_confidence


import os
import numpy as np # type: ignore
import pandas as pd # type: ignore

from .comm import smartPrint, tidyMessage
from .org import isFileOrFolder 
from .utils import flattenList

from scipy.stats import norm # type: ignore


def main(args):

	verbose = args.verbose
	directory = args.input
	overwrite = args.over_write
	confidence = float(args.confidence) / 100.0
	z_value = (1-(1-confidence)/2)
	add_noise = args.include_noise

	msg = 'AMiGA is parsing your file(s)'

	smartPrint('',verbose)
	smartPrint(tidyMessage(msg),verbose)

	directory,filename = isFileOrFolder(directory,up=1)

	# packge filename(s) into a list
	if filename:
		ls_files = [f'{directory}{os.sep}{filename}']
	else:
		ls_files = filename

	for lf in ls_files:

		df = pd.read_csv(lf,sep='\t',header=0,index_col=0)

		# define file name for the updated dataframe
		if overwrite: 
			new_name = lf
		elif lf.endswith('.txt'):
			new_name = f'{lf[:-4]}_confidence.txt'
		else:
			new_name = f'{lf}_confidence.txt'

		# compute confidecne intervals and save results
		if args.type == 'Parameters':
			df = get_parameter_confidence(df,z_value)
			df.to_csv(new_name,sep='\t',header=True,index=True)
		elif args.type == 'Curves':
			df = get_curve_confidence(df,z_value,add_noise)
			df.to_csv(new_name,sep='\t',header=True,index=False)


def get_parameter_confidence(df,z_value):
	''''Compute confidence intervals for parameters tabulated in dataframe

	Args:
		df (pandas.DataFrame): arameters must have both mean({}) and std({}) values
		confidence (float)
		add_noise (boolean): include measurement variance when computing confidence intervals

	Returns:
		df (pandas.DataFrame)
	'''

	def conf_interval(mu_std,z_value=0.975):
		mean,std = mu_std
		low = mean-norm.ppf(z_value)*std
		upp = mean+norm.ppf(z_value)*std
		return [low,upp]
	
	def getp(param):
		headers = [f"{summary}({param})" for summary in ['mean','std']]
		return row[headers].values
	
	def stats_p(param):
		headers = [f"{summary}({param})" for summary in ['mean','std','low','upp']]
		return headers

	# get all params in table with both mean and std
	keys = df.keys()
	set1 = [ii.split('mean(')[1][:-1] for ii in keys if ii.startswith('mean(')]
	set2 = [ii.split('std(')[1][:-1] for ii in keys if ii.startswith('std(')]
	params = sorted(list(set(set1).intersection(set2)))

    # define new dataframe for storing lower and upper confidnce intervals
	ci_keys =flattenList([[f'low({ii})',f'upp({ii})'] for ii in params])
	ci_df = pd.DataFrame(index=df.index,columns=ci_keys)

    # for each sample/row, compute and store confidence intervals
	for idx,row in df.iterrows():
		ci_df.loc[idx,ci_keys] = flattenList([conf_interval(getp(ii),z_value) for ii in params])

	# order columns by grouping stats of each parameter
	order_params = (flattenList([stats_p(pp) for pp in params]))
	non_params = list(set(keys).difference(set(order_params)))

	# put together the dataframe
	df = df.join(ci_df).loc[:,non_params+order_params]

	return df

def get_curve_confidence(df,z_value,add_noise=False):
	'''Compute confidence intervals for growth and growth rate functions

	Args:
		df (pandas.DataFrame): columns must include mu, Sigma, mu1, Sigma1, Noise
		confidence (float)
		add_noise (boolean): include measurement variance when computing confidence intervals

	Returns:
		df (pandas.DataFrame)
	'''

	def conf_interval(mu,sigma,noise,z_value,add_noise=add_noise):
		if add_noise:
			sigma = sigma + noise
		else:
			sigma = sigma
		band = norm.ppf(z_value)*np.sqrt(sigma)
		low = list((mu-band).values)
		upp = list((mu+band).values)
		return [low,upp]

    # what are original keys & create df for storing intervals
	keys = df.keys()
	ci_df = pd.DataFrame(index=df.index,columns=['low','upp','low1','upp1'])

    # get confidence intervals for grwoth and growth rate
	foo0 = conf_interval(df['mu'],df['Sigma'],df['Noise'],z_value=z_value,add_noise=add_noise)
	foo1 = conf_interval(df['mu1'],df['Sigma1'],None,z_value=z_value,add_noise=False)

    # store in dataframe
	ci_df.loc[:,'low'] = foo0[0]
	ci_df.loc[:,'upp'] = foo0[1]
	ci_df.loc[:,'low1'] = foo1[0]
	ci_df.loc[:,'upp1'] = foo1[1]

    # reorder parameters
	order_params = ['mu','low','upp','Sigma','mu1','low1','upp1','Sigma1','Noise']
	non_params = list(set(keys).difference(set(order_params)))

	# put together the dataframe
	df = df.join(ci_df).loc[:,non_params+order_params]

	return df




#!/usr/bin/env python

'''
AMiGA library of functions for reading and cleaning data.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (8 functions)

# readPlateReaderFolder
# findPlateReaderFiles
# breakDownFilePath
# readPlateReaderData
# checkFileEncoding
# findRowsAndIndex
# isWellId
# listTimePoints

import os
import sys
import numpy as np # type: ignore
import pandas as pd # type: ignore

from string import ascii_uppercase

from .config import config
from .detail import parseWellLayout
from .comm import smartPrint
from .org import assemblePath


def readPlateReaderFolder(filename=None,directory=None,interval=dict(),save=False,verbose=False):
    '''
    Finds, reads, and formats all files in a directory to be AMiGA-compatible.

    Args:
        filename (str or None): 
            if str: path to a single data file to be read.
            if None: user is interested in reading one or more data files (so user must pass directory argument).
        directory (dictionary or str or None):
            if dictionary: Keys are folder names, values are their paths. Keys must include 'data' and 'derived'
                'data' sub-folder must exist and house one or more data files to be read. 
            if str: path to folder that houses one ore more data files to be read.
            if None: user is interested in reading a single data file (so user must pass filename argument).
        interval (dictionary or numeric):
            if numeric: must be int or float.
            if dictionary: Keys are file names, values are their respective interval parameter, e.g. 
                {'CD2015_PM1-1':600,'CD2048_PM1-1':900}). If a filename does not have a corresponding key in the
                dictionary, the default parameter for 'interval' in the 'config' dictionary will be used. 
        save (boolean): will save AMiGA-formatted file in the 'derived' or input folder as a TSV file.
        verbose (boolean)
    '''

    if (filename is None) and (directory is None): 
        sys.exit('FATAL USER ERROR: User must pass either a filename or a directory argument')

    # what is the data folder (folderpath) and where to save formatted data (copydirectory)? 
    if isinstance(directory,dict):
        folderpath = directory['data']
        copydirectory = directory['derived']
    elif isinstance(directory,str):
        copydirectory = folderpath = directory
    elif directory is None:
        copydirectory = folderpath = os.path.dirname(filename)

    # user may have passed a specific file or a directory to the input argument
    if filename:
        filepaths = [f'{folderpath}{os.sep}{filename}']
    else:
        filepaths = findPlateReaderFiles(folderpath)
    # either way, filepaths must be an iterable list or array

    # read one data file at a time
    df_dict = {}
    for filepath in sorted(filepaths):

        # communicate with user
        smartPrint(f'Reading {filepath}',verbose)

        # get extension-free file name and path for derived copy
        _, filebase, newfilepath = breakDownFilePath(filepath,copydirectory=copydirectory)

        # set the interval time
        if isinstance(interval,(int,float)):
            plate_interval = float(interval)
        elif filebase in interval.keys():
            plate_interval = interval[filebase]
        else: 
            plate_interval = config['interval']

        # read and adjust file to format: time by wells where first column is time and rest are ODs
        df = readPlateReaderData(filepath,plate_interval,copydirectory,save=save)
        df_dict[filebase] = df#..iloc[nskip:,:]

    smartPrint('',verbose)  # print empty newline, for visual asethetics only

    return df_dict


def findPlateReaderFiles(directory,extension=None):
    '''
    Recrusivelys searches a directory for all files with specific extensions.

    Args:
        directory (str): path to data directory

    Returns:
        list_files (list): list of of paths to data files
    '''

    # you can modify this to include other extensions, but internal data must still be tab-delimited 
    if extension is None:
        extension = ('.txt','TXT','tsv','TSV','asc','ASC')

    # recursively walk through a directory, if nested
    list_files = []

    for (dirpath,dirnames,filenames) in os.walk(directory):

        # only keep files with acceptable extensions
        filenames = [ii for ii in filenames if ii.endswith(extension)]

        # compose and store filepaths but avoid double slashes (i.e. //) between directory names
        for filename in filenames:

            list_files.append(assemblePath(dirpath,filename))
  
    return list_files


def breakDownFilePath(filepath,copydirectory):
    '''
    Breaks down a file path into several components.

    Args:
        filepath (str)
        save_dirname (str): directory where a copy of the file would be stored

    Returns:
        filename (str): basename without path
        filebase (str): basename without path and without extension
        newfilepath (str): filename with path and with extension repalced to .tsv

    Example input: '/Users/firasmidani/RandomFileName.asc' will return
        filename --> RandomFileName.asc
        filebase --> RandomFileName
        newfilepath --> /Users/firasmiani/RandomFileName.tsv  
    '''

    filename = os.path.basename(filepath)
    filebase = '.'.join(filename.split('.')[:-1])

    newfilepath = assemblePath(copydirectory,filebase,'.tsv')

    return filename, filebase, newfilepath
    

def readPlateReaderData(filepath,interval,copydirectory,save=False):
    '''
    Reads a single file and adjusts it to AMiGA-desired format. The desired format is 
        time point (row) by well (col, where first column is actual time point, so row
        names are arbitray numerical order (1...n). All cells should be floats.

    Args:
        filepath (str)
        interval (int): time interval length
        copydirectory (str): location  
    '''

    # initialize useful derivatives of filepath
    filename,filebase,newfilepath = breakDownFilePath(filepath,copydirectory)

    # make sure file is ASCII- or BOM-encoded
    encoding = checkFileEncoding(filepath)

    # identify number of rows to skip and presence/location of index column (i.e. row names)
    skiprows,skipfooter,index_col = findRowsAndIndex(filepath,encoding=encoding)

    # read tab-delimited data file
    df = pd.read_csv(filepath,sep='\t',header=None,index_col=index_col,skiprows=skiprows,skipfooter=skipfooter,encoding=encoding,engine='python')

    # explicitly define time series based on data size and time interval 
    df.columns = listTimePoints(interval=interval,numTimePoints=df.shape[1])

    # if index column is absent, create one 
    if index_col is None:
        nrows = df.shape[0]
        df.index = parseWellLayout(order_axis=0).index[0:nrows].values

    # explicilty assign column names 
    df.T.index.name = 'Time'

    # remove columns (time points) with only NA values (sometimes happens in plate reader files)
    df = df.iloc[:,np.where(df.notna().all(axis=0))[0]]

    # remove rows (smples) with only NA values (happens if there is meta-data in file after measurements)
    df = df.iloc[np.where(df.T.notna().all(axis=0))[0],:]

    # strip leading zeros in row or well IDs
    df.index = [ii[0] + ii[1:].lstrip('0') for ii in df.index]

    try:
        df = df.astype(float)
    except Exception:
        msg = f'\nFATAL DATA ERROR: {filename} data file is not properly formatted. '
        msg += 'AMiGA expected measurement values that are numerical integers or floats '
        msg += 'but instead detected strings (text characters). '
        msg += 'Please see documentation for instructions on how to properly format data files. \n'
        sys.exit(msg)

    # set to following format: time point (row) by well (column)
    df = df.T.reset_index(drop=False) # values are OD (float) except first column is time (float)

    # save derived data copy in proper location
    if save:
        df.to_csv(newfilepath,sep='\t',header=True)  # does it save header index name (i.e. Time)

    return df


def checkFileEncoding(filepath):
    '''
    Identifies the correct file encoding needed for python open() to read a text file. 
        It does this brutely using error-detection. If none of the pre-accepted encondings
        are detected, AMiGA will prematurely terminate with a descriptive message to user. 

    Args:
        filepath (str)

    Returns:
        encodign (str): see limited options below. 
    '''

    # not sure if ASICC is necesary since ASCII is a subset of UTF-8
    for encoding in ['UTF-8','UTF-16','UTF-32','ASCII']:
        try:
            open(filepath,encoding=encoding).readline()
            return encoding
        except UnicodeDecodeError:
            pass

    # exit 
    msg = f'FATAL DATA ERROR: AMiGA cannot read{filepath}. '
    msg += 'AMiGA can only read data text files encoded with either '
    msg += 'UTF-8, UTF-16, UTF-32, or ASCII.\n'
    sys.exit(msg) 
 

def findRowsAndIndex(filepath,encoding):
    '''
    Searches for line that begins with a Well ID (defined as a letter followed by digits), 
        determines the number of rows that need to be skipped for reading this line, and
        indicates if index column was not found. Then, reverse reads file and searches for
        line that begins with a Well ID to detemine the number of lines to skip in footer. 

    Args:
        filepath (str)

    Returns:
        firstrow (int): number of lines that need to be skipped to read data for first well
        lasttrow (int): number of lines at the bottom that need to be skipped to read data to step at last ro
        index_column (0 or None): location of row namess, if found as first character (0) or not found (None) 
    '''

    fid = open(filepath,encoding=encoding)

    firstrow = 0
    for line in fid.readlines():
        line_start = line.strip().split('\t')[0]
        if isWellId(line_start):
            index_column = 0  # row names are the zero-indexed column
            break
        firstrow += 1
    else:
        firstrow = 0
        lastrow = 0
        index_column = None  # row names were not found

    fid.close()

    # I could have tweaked the above for-else statement to detect the footer using the same iterator, BUT
    #     that would have required reading eahc line of the file. If the file is very large, this can 
    #     increase time. Instead, I simply reverse the iterated lines and stop reading once Well ID is detected.

    fid = open(filepath,encoding=encoding)

    lastrow = 0
    for line in reversed(fid.readlines()):
        line_start = line.strip().split('\t')[0]
        if isWellId(line_start):
            break
        lastrow += 1

    fid.close()

    return firstrow,lastrow,index_column


def isWellId(item):
    '''
    Checks if argument is string and if it has a letter for first character and digits for remaining characters.

    Args:
        item (string)

    Returns:
        (boolean)
    '''

    if not isinstance(item,str):
        return False

    if len(item) < 2:
        return False

    if (item.lstrip('0')[0] in ascii_uppercase) and (item[1:].isdigit()):
        return True
    
    return False


def listTimePoints(interval,numTimePoints):
    '''
    Constructs a numpy.array of a time series based on time interval length and number of time points.

    Args:
        interval (int or float, latter preferred)
        numTimePoints (int)

    Returns:
        time_series (numpy.array of floats)
    '''

    time_series = np.arange(start=0.0,stop=interval*numTimePoints,step=interval)

    return time_series




#!/usr/bin/env python

'''
AMiGA wrapper: main driver script for Analysis of Microbial Growth Assays (AMiGA).
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (1 class with 4 class functions)

# Command
# __init__
# summarize
# test
# fit

from .analyze import basicSummaryOnly, runGrowthFitting
from .comm import tidyMessage
from .config import config
from .detail import assembleMappings
from .interface import interpretParameters
from .org import isFileOrFolder, mapDirectories, mapFiles, validateDirectories
from .read import readPlateReaderFolder
from .test import HypothesisTest
from .trim import trimInput, flagWells


class Command:

    def __init__(self,args,config=config):

        # parse terminal command for arguments
        #args = parseCommand(args,config);

        # did the user provide a path that points to a file or directory?
        parent,filename = isFileOrFolder(args.input) 

        # communicate with user
        print()
        print(tidyMessage('AMiGA is peeking inside the working directory'))

        # define file paths for AMiGA-relevant folders (dict)
        directory = mapDirectories(parent)

        # define file paths for AMiGA-relevant files (dict)
        files = mapFiles(directory)

        # validate working directory structure and contents (no output)
        validateDirectories(directory,verbose=args.verbose)

        # communicate with user
        print(tidyMessage('AMiGA is parsing command-line arguments and parameter files'))

        # interpret terminal command or text files for parameters (dict)
        params,args = interpretParameters(files,args,verbose=args.verbose)

        # communicate with user
        print(tidyMessage('AMiGA is parsing and cleaning data files'))

        # parse data files (dict)
        data = readPlateReaderFolder(filename,directory,interval=params['interval'],
            save=args.save_cleaned_data,verbose=args.verbose)

        # communicate with user
        print(tidyMessage('AMiGA is parsing and reading mapping files'))

        # parse mapping files and mapping data (dict)
        mappings = assembleMappings(data,directory['mapping'],files['meta'],
            save=args.save_mapping_tables,verbose=args.verbose)

        self.directory = directory
        self.params = params
        self.args = args
        self.data = data
        self.mappings = mappings

    def summarize(self):

        # flag wells
        self.mappings = flagWells(self.mappings,self.params['flag'],verbose=self.args.verbose,drop=False)

        # communicate with user
        print(tidyMessage('AMiGA is summarizing and plotting data'))

        # plot and summarize plate(s), if it is the only request
        basicSummaryOnly(self.data,self.mappings,self.directory,self.args,verbose=self.args.verbose)


    def test(self):

        # communicate with user
        print(tidyMessage('AMiGA is testing data based on user input'))

        # run hypothesis testing, if requested
        HypothesisTest(self.mappings,self.data,self.params,self.args,self.directory)


    def fit(self):

        args = self.args

        # communicate with user
        print(tidyMessage('AMiGA is preparing data based on user input'))

        # trim mapping data based on user input
        data,mappings = trimInput(self.data,self.mappings,self.params,nskip=args.skip_first_n,verbose=args.verbose)

        # communicate with user
        print(tidyMessage('AMiGA is fitting growth curves'))

        # run growth fitting 
        runGrowthFitting(data,mappings,self.directory,args,verbose=args.verbose)




#!/usr/bin/env python

'''
AMiGA library for the Gaussian Process Model class.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"

# TABLE OF CONTENTS (1 function and 1 class with 7 sub-functions)

# describeVariance
# 
# Model (CLASS)
#   __init__
#   permute
#   fit
#   predict_y0
#   predict_y1
#   predict_y2
#   run

import warnings
import numpy as np # type: ignore
import pandas as pd # type: ignore

from GPy.models import GPRegression # type: ignore

from scipy.ndimage import filters # type: ignore

from .kernel import buildKernel,addFixedKernel
from .curve import GrowthCurve
from .utils import uniqueRandomString, subsetDf, getValue

if getValue('Ignore_RuntimeWarning'):
    warnings.filterwarnings("ignore", category=RuntimeWarning) 


def describeVariance(df,time='X0',od='Y'):
    '''
    df columns ['X0','X1',...,'Y']
    values of Xs except fo X0 should be non-unique
    '''

    window = getValue('variance_smoothing_window')
    nX = len(df[time].drop_duplicates())
    if window < 1:
        window = int(np.ceil(nX*window))
    
    df = df.sort_values(time)
    df.reset_index(drop=True,inplace=True)

    error = df[[time,'OD']]
    error = error.groupby([time]).apply(lambda x: np.nanvar(x.OD))
    error = pd.DataFrame(error,columns = ['error'])
    error = error.reset_index()
    error = error.drop_duplicates().set_index(time).sort_index()
    error.loc[:,'error'] = filters.gaussian_filter1d(error.error.values,window)
    df = pd.merge(df,error,on=time,how='outer').sort_values([time])
    
    return df


class GrowthModel:


    def __init__(self,df=None,model=None,x_new=None,baseline=1.0,ARD=False,heteroscedastic=False,nthin=1,logged=True):
        '''
        Data structure for Gaussian Process regression and related parameter inference.

        Attributes:
            x (numpy.ndarray): independent variables (N x D), where N is the number of observations, and 
                D is the number of dimensions (or variables).
            y (numpy.ndarray): dependent variables (N x 1), where N is the number of observations, and 
                the only column is the dependent or obesrved variable (often Optical Density or OD).
            key (dict or pandas.DataFrame): dictionary (k) or pandas.DataFrame (1 x k) that describes 
                k experimental variables about sample. Must include 'OD_Baseline' and 'Fold_Change' variables.

        Notes:
            for growth curve analysis, it is assumed that y was log-transformeed and baseline-corrected. 

        '''

        if model:
            self.model = model
            self.x_new = x_new
            self.ARD = ARD
            self.baseline = baseline
            self.logged = logged
            self.y = None
            self.df = None
            return None

        self.df = df.copy()

        # create a dummy non-unique variable/column
        foo = uniqueRandomString(avoid=df.keys())
        df[foo] = [1]*df.shape[0]
        varbs = df.drop(labels=['Time','OD'],axis=1).drop_duplicates()

        # for each unique non-time variable, estimate variance
        new_df = []
        for idx,row in varbs.iterrows():
            sub_df = subsetDf(df,row.to_dict())
            sub_df = describeVariance(sub_df,time='Time',od='OD')
            new_df.append(sub_df)

        new_df = pd.concat(new_df,axis=0)
        new_df = new_df.drop(labels=[foo],axis=1)

        # construct a thinner dataframe to speed up regression
        time = new_df.Time.sort_values().unique()
        time = time[::int(nthin)]
        thin_df = new_df[new_df.Time.isin(time)]
        
        # predictions of error and new are based on full dataframe
        tmp = new_df.drop(labels=['OD'],axis=1).drop_duplicates()
        error_new = tmp.loc[:,['error']].values
        x_new = tmp.drop(labels=['error'],axis=1).values

        # regression are based on input/output/error from thinned dataframe
        x = thin_df.drop(labels=['OD','error'],axis=1).values
        y = thin_df.loc[:,['OD']].values
        error = thin_df.loc[:,['error']].values
        x_keys = thin_df.drop(labels=['OD','error'],axis=1).keys()

        # save attributes
        self.x_keys = x_keys
        self.x_new = x_new
        self.x = x
        self.y = y 
        self.error = error
        self.error_new = error_new
        self.baseline = baseline
        self.logged = logged

        self.model = model
        self.ARD = ARD
        self.heteroscedastic = heteroscedastic
        self.noise = None


    def permute(self,varb=None):

        # get model input
        x = self.x.copy()
        y = self.y.copy()

        # shuffle targeet variable
        col = np.where(varb==self.x_keys)[0] # which one?
        shuffled = np.random.choice(x[:,col].ravel(),size=x.shape[0],replace=False)
        x[:,col] = shuffled[:,np.newaxis] # replace

        # same steps as fit, below can be done more concisely, but I worry about minor differenes
        x_dim = x.shape[1]
        y_dim = y.shape[1]

        kern = buildKernel(x_dim,x,y,ARD=self.ARD)
        mcopy = GPRegression(x,y,kern)
  
        if self.heteroscedastic:
            kern = addFixedKernel(kern,y_dim,self.error)
            mcopy = GPRegression(x,y,kern)
     
        mcopy.optimize()

        return mcopy.log_likelihood()


    def fit(self):

        if self.model:
            self.noise = self.model.Gaussian_noise.variance[0]
            return None

        x_dim = self.x.shape[1] # number of input dimensions, 1 if only time
        y_dim = self.y.shape[1] # number of ouptut dimensions, typically only 1 for log OD

        kern = buildKernel(x_dim,self.x,self.y,ARD=self.ARD)
        m = GPRegression(self.x,self.y,kern)

        if self.heteroscedastic:
            kern = addFixedKernel(kern,y_dim,self.error)
            m = GPRegression(self.x,self.y,kern)

        m.optimize()

        self.noise = m.Gaussian_noise.variance[0] # should be negligible (<1e-10) for full model

        if self.heteroscedastic:
            m.kern = m.kern.parts[0] # cannot predict with fixed kernel, so remove it

        self.model = m


    def predict_y0(self):

        x = self.x_new
        m = self.model

        mu,cov = m.predict(x,full_cov=True,include_likelihood=False)

        self.y0 = mu
        self.cov0 = cov

    def predict_y1(self):

        x = self.x_new
        m = self.model

        ndim = x.shape[1]

        if (ndim > 1) and (not self.ARD):
            mu,cov = (m.predictive_gradients(x)[0],None)
        else:
            mu,cov = m.predict_jacobian(x,full_cov=True)
            cov = cov[:,:,0,0]

        self.y1 = mu[:,0]
        self.cov1 = cov


    def predict_y2(self):

        x = self.x_new
        y = self.y1

        m = GPRegression(x,y)
        m.optimize()
        mu,cov = (m.predictive_gradients(x)[0],None)

        self.y2 = mu[:,0]
        self.cov2 = cov


    def run(self,simple=False,name=None,predict=True):

        self.fit()

        if not predict:
            return self, self.model.log_likelihood()

        self.predict_y0()
        self.predict_y1()
        self.predict_y2()

        if self.df is not None:
            actual_input = self.df.OD.values[:,np.newaxis]
        else:
            actual_input = None

        if not simple:
            time = self.x_new[:,0][:,np.newaxis]
            curve = GrowthCurve(x=time,y=actual_input,y0=self.y0,y1=self.y1,y2=self.y2,
                                cov0=self.cov0,cov1=self.cov1,
                                baseline=self.baseline,name=name,logged=self.logged)
            return curve



#!/usr/bin/env python

'''
AMiGA library for the AMiGA class for parsing command-line arguments. 
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (1 class with 8 sub-functions and 1 auxiliary function)

# AMiGA (CLASS)
#   __init__
#   compare
#   fit
#   heatmap
#   normalize
#   print_defaults
#   summarize
#   test
#
# print_arguments

import os
import sys
import argparse
import importlib.metadata


# Add the parent directory to sys.path to allow package imports
sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))

from amiga.libs.commands import Command
from amiga.libs.comm import tidyMessage, tidyDictPrint
from amiga.libs.config import config

# define long string variables

usage_init = '''amiga <command> [<args>]

The most commnly used amiga commands are:
    summarize       Perform basic summary and plot curves
    fit             Fit growth curves
    normalize       Normalize growth parameters of fitted curves
    compare         Compare summary statistics for two growth curves
    test            Test a specific hypothesis
    heatmap         Plot a heatmap
    get_confidence  Compute confidence intervals for parameters or curves
    get_time        Get time at which growth reaches a certain value
    print_defaults  Shows the default values stored in libs/config.py

See `amiga <command> --help` for information on a specific command.
For full documentation, see https://firasmidani.github.io/amiga
'''

# define auxiliary functions

def print_arguments(args):

    msg = '\n'
    msg += tidyMessage('User provided the following command-line arguments:')
    msg += '\n'
    msg += tidyDictPrint(vars(args))

    print(msg)


class AMiGA:
    '''
    Class for interpreting the arguments passed by the user to AMiGA. If the verobse argument 
        is set to True, sub-commands will also print a message summarizing the 
        the user-passed command-line arguments.
    '''

    def __init__(self):
        parser = argparse.ArgumentParser(
            usage=usage_init)
        parser.add_argument('command', help='Subcommand to run. See amiga --help for more details.')
        parser.add_argument('-v','--version', action="version", version=f"amiga {importlib.metadata.version('amiga')}", help="Show version and exit.")

        args = parser.parse_args(sys.argv[1:2])

        if not hasattr(self,args.command):
            print(f"Error: Unknown command '{args.command}'\n")
            parser.print_help()
            exit(1)
        else: 
            getattr(self, args.command)()

        # Prevent Python from returning the object representation
        sys.exit(0)

    def compare(self):

        parser = argparse.ArgumentParser(
            description='Compare two growth curves')

        parser.add_argument('-i','--input',required=True,action='append')
        parser.add_argument('-o','--output',required=True,help='ouptut filename including path')
        parser.add_argument('-s','--subset',required=True)
        parser.add_argument('--confidence',required=False,type=float,default=95,
            help='Must be between 80 and 100. Default is 95.')
        parser.add_argument('--verbose',action='store_true',default=False)

        if len(sys.argv) == 2:
            parser.print_help()
            sys.exit(0)

        args = parser.parse_args(sys.argv[2:])

        if args.verbose:
            print_arguments(args)

        if (args.confidence < 80) | (args.confidence > 100):
            msg = 'FATAL USER ERROR: Confdience must be between 80 and 100.'
            sys.exit(msg)
        
        from amiga.libs.compare import main as compare_main

        compare_main(args)


    def get_confidence(self):

        parser = argparse.ArgumentParser(
            description = 'Compute confidence intervals for parameters or curves.')
        parser.add_argument('-i','--input',required=True)
        parser.add_argument('--type',required=True,default='Parameters',choices=['Parameters','Curves'])
        parser.add_argument('--confidence',required=False,type=float,default=95,
            help='Must be between 80 and 100. Default is 95.')
        parser.add_argument('--include-noise',action='store_true',default=False,
            help='Include the estimated measurement noise when computing confidence interval (For Curves Only).')
        parser.add_argument('--over-write',action='store_true',default=False,
            help='Over-write file otherwise a new copy is made with "_confidence" suffix')
        parser.add_argument('--verbose',action='store_true',default=False)

        if len(sys.argv) == 2:
            parser.print_help()
            sys.exit(0)

        args = parser.parse_args(sys.argv[2:])

        if args.verbose:
            print_arguments(args)

        if (args.confidence < 80) | (args.confidence > 100):
            msg = 'FATAL USER ERROR: Confdience must be between 80 and 100.'
            sys.exit(msg)
        
        from amiga.libs.confidence import main as get_confidence_main
        
        get_confidence_main(args)


    def get_time(self):

        parser = argparse.ArgumentParser(
            description='Get time at which OD reaches a certain value')

        parser.add_argument('--gp-data',required=True)
        parser.add_argument('--summary',required=True)
        parser.add_argument('--threshold',required=True,type=float)
        parser.add_argument('--curve-format',required=False,default='OD_Growth_Fit',
            choices=['OD_Data','OD_Fit','GP_Input','GP_Output',
            'OD_Growth_Fit','OD_Growth_Data','GP_Derivative'])

        if len(sys.argv) == 2:
            parser.print_help()
            sys.exit(0)

        args = parser.parse_args(sys.argv[2:])

        from amiga.libs.thresholds import get_time_main

        get_time_main(args)


    def fit(self):

        parser = argparse.ArgumentParser(
            description='Fit growth curves')

        # defining input/output
        parser.add_argument('-i','--input',required=True)
        parser.add_argument('-o','--output',required=False)
        parser.add_argument('-f','--flag',required=False)
        parser.add_argument('-s','--subset',required=False)
        parser.add_argument('-t','--interval',required=False)
        parser.add_argument('-tss','--time-step-size',action='store',type=int,default=1)#11
        parser.add_argument('-sfn','--skip-first-n',action='store',type=int,default=0)
        parser.add_argument('--do-not-log-transform',action='store_true',default=False)
        parser.add_argument('--subtract-blanks',action='store_true',default=False)
        parser.add_argument('--subtract-control',action='store_true',default=False)
        parser.add_argument('--keep-missing-time-points',action='store_true',default=False)
        parser.add_argument('--verbose',action='store_true',default=False)
        parser.add_argument('--plot',action='store_true',default=False)
        parser.add_argument('--plot-derivative',action='store_true',default=False)
        parser.add_argument('--pool-by',required=False)
        parser.add_argument('--save-cleaned-data',action='store_true',default=False)
        parser.add_argument('--save-mapping-tables',action='store_true',default=False)
        parser.add_argument('--save-gp-data',action='store_true',default=False)
        parser.add_argument('--merge-summary',action='store_true',default=False)
        parser.add_argument('--fix-noise',action='store_true',default=False)
        parser.add_argument('--sample-posterior',action='store_true',default=False)

        if len(sys.argv) == 2:
            parser.print_help()
            sys.exit(0)

        args = parser.parse_args(sys.argv[2:])

        if (args.fix_noise) and (not args.pool_by):
            msg = '\nWARNING: --fix-noise and is only applicable if user also '
            msg += 'requests pooling with the --pool-by argument.'
            print(msg)

        if (args.sample_posterior) and (not args.pool_by):
            msg = '\nWARNING: --sample-posterior is only applicable if user also '
            msg += 'requests pooling with the --pool-by argument.'
            print(msg)

        if args.do_not_log_transform:
            args.log_transform = False
        else:
            args.log_transform = True

        if args.subset:
            args.merges = True

        if args.verbose:
            print_arguments(args)

        # unnecessary args
        args.hypothesis = None

        Command(args).fit()


    def heatmap(self):

        parser = argparse.ArgumentParser(
            description='Plot a heatmap')
        parser.add_argument('-i','--input',required=True)
        parser.add_argument('-o','--output',required=True)
        parser.add_argument('-s','--subset',required=False)
        parser.add_argument('-v','--value',required=True)
        parser.add_argument('-x','--x-variable',required=True)
        parser.add_argument('-y','--y-variable',required=True)
        parser.add_argument('-p','--operation',required=False,default='mean',
            choices=['mean','median'])
        parser.add_argument('-f','--filter',required=False)
        parser.add_argument('-t','--title',required=False)
        parser.add_argument('--kwargs',required=False)
        parser.add_argument('--verbose',action='store_true',default=False)
        parser.add_argument('--save-filtered-table',action='store_true',default=False)

        # sizing 
        parser.add_argument('--width-height',required=False,nargs=2)
        parser.add_argument('--colorbar-ratio',required=False,default=0.1,type=float,
            help='Proportion of figure size devoted to color bar. Default is 0.1')

        # coloring labels
        parser.add_argument('--color-x-by',required=False)
        parser.add_argument('--color-y-by',required=False)
        parser.add_argument('--color-file-x',required=False)
        parser.add_argument('--color-file-y',required=False)
        parser.add_argument('--color-scheme-x',required=False)
        parser.add_argument('--color-scheme-y',required=False)
        parser.add_argument('--color-x-ratio',required=False,default=0.1,type=float,
            help='Proportion of the heatmap devoted to the column color labels. Default is 0.1')
        parser.add_argument('--color-y-ratio',required=False,default=0.1,type=float,
            help='Proportion of the heatmap devoted to the row color labels. Default is 0.1')
        parser.add_argument('--missing-color',required=False,default=None)

        # sorting
        parser.add_argument('--cluster-x',required=False,default=False,action='store_true')
        parser.add_argument('--cluster-y',required=False,default=False,action='store_true')     
        parser.add_argument('--sort-x-by',required=False)
        parser.add_argument('--sort-y-by',required=False)

        # handling missing data
        parser.add_argument('--keep-rows-missing-data',action='store_true',default=False,
            help='Drops columsn that have any missing data')
        parser.add_argument('--keep-columns-missing-data',action='store_true',default=False,
            help='Drops rows that have any missing data')

        # handling label sizes and adjusting tick labels
        parser.add_argument('--x-tick-labels-scale',required=False,default=config['x_tick_labels_scale'],
            help='Must be between 0 (smallest) and 1 (largest).')
        parser.add_argument('--y-tick-labels-scale',required=False,default=config['y_tick_labels_scale'],
            help='Must be between 0 (smallest) and 1 (largest).')
        parser.add_argument('--color-bar-labels-scale',required=False,default=config['color_bar_labels_scale'],
            help='Must be between 0 (smallest) and 1 (largest).')
        parser.add_argument('--x-rotation',required=False,default=90)
        parser.add_argument('--highlight-labels',required=False)

        if len(sys.argv) == 2:
            parser.print_help()
            sys.exit(0)

        args = parser.parse_args(sys.argv[2:])

        if args.verbose:
            print_arguments(args)
            
        from amiga.libs.heatmap import main as heatmap_main
        
        heatmap_main(args)


    def normalize(self):
        parser = argparse.ArgumentParser(
            description='Compare two growth curves')

        parser.add_argument('-i','--input',required=True)
        parser.add_argument('--over-write',action='store_true',default=False,
            help='Over-write file otherwise a new copy is made with "_normalize" suffix')
        parser.add_argument('--verbose',action='store_true',default=False)
        parser.add_argument('--group-by',required=False)
        parser.add_argument('--normalize-by',required=False)
        parser.add_argument('--normalize-method',action='store',default='subtraction',
            choices=['division','subtraction'])

        if len(sys.argv) == 2:
            parser.print_help()
            sys.exit(0)

        args = parser.parse_args(sys.argv[2:])

        if args.verbose:
            print_arguments(args)

        from amiga.libs.normalize import main as normalize_main

        normalize_main(args)


    def print_defaults(self):

        msg = '\nDefault settings for select variables. '
        msg += 'You can adjust these values in "amiga/libs/config.py". \n\n'
        msg += tidyDictPrint(config)
        print(msg)
        #sys.exit(msg)


    def summarize(self):
        parser = argparse.ArgumentParser(
            description='Perform a basic summary and plot curves')

        parser.add_argument('-i','--input',required=True)
        parser.add_argument('-o','--output',required=False)
        parser.add_argument('--dont-plot',action='store_true',default=False)
        parser.add_argument('--merge-summary',action='store_true',default=False)
        parser.add_argument('--verbose',action='store_true',default=False)
        parser.add_argument('-f','--flag',required=False)
        parser.add_argument('-s','--subset',required=False)
        parser.add_argument('-y','--hypothesis',required=False)
        parser.add_argument('-t','--interval',required=False)
        parser.add_argument('--save-cleaned-data',action='store_true',default=False)
        parser.add_argument('--save-mapping-tables',action='store_true',default=False)
        parser.add_argument('--subtract-blanks',action='store_true',default=False)
        parser.add_argument('--subtract-control',action='store_true',default=False)

        if len(sys.argv) == 2:
            parser.print_help()
            sys.exit(0)

        args = parser.parse_args(sys.argv[2:])

        args.subset = None  ## subsetting is not implemented
        args.hypothesis = None ## hypothis are not implemented

        if args.verbose:
            print_arguments(args)

        Command(args).summarize()


    def test(self):
        parser = argparse.ArgumentParser(
            description='Test for differential growth between two conditions')

        # defining input/output
        parser.add_argument('-i','--input',required=True)
        parser.add_argument('-o','--output',required=False)
        parser.add_argument('-f','--flag',required=False)
        parser.add_argument('-s','--subset',required=False)
        parser.add_argument('-t','--interval',required=False)
        parser.add_argument('-y','--hypothesis',required=True)
        parser.add_argument('-sfn','--skip-first-n',action='store',type=int,default=0)
        parser.add_argument('-tss','--time-step-size',action='store',type=int,default=1)#11
        parser.add_argument('-np','--number-permutations',action='store',type=int,default=0)
        parser.add_argument('-fdr','--false-discovery-rate',action='store',type=int,default=10)
        parser.add_argument('--confidence',required=False,type=float,default=95,
            help='Must be between 80 and 100. Default is 95.')
        parser.add_argument('--subtract-blanks',action='store_true',default=False)
        parser.add_argument('--subtract-control',action='store_true',default=False)
        parser.add_argument('--verbose',action='store_true',default=False)
        parser.add_argument('--fix-noise',action='store_true',default=False)
        parser.add_argument('--include-gaussian-noise',action='store_true',default=False)
        parser.add_argument('--sample-posterior',action='store_true',default=False)
        parser.add_argument('--dont-plot',action='store_true',default=False)
        parser.add_argument('--dont-plot-delta-od',action='store_true',default=True)
        parser.add_argument('--save-cleaned-data',action='store_true',default=False)
        parser.add_argument('--save-mapping-tables',action='store_true',default=False)
        parser.add_argument('--save-gp-data',action='store_true',default=False)
        parser.add_argument('--merge-summary',action='store_true',default=False)
        parser.add_argument('--do-not-log-transform',action='store_true',default=True)

        if len(sys.argv) == 2:
            parser.print_help()
            sys.exit(0)

        args = parser.parse_args(sys.argv[2:])

        if (args.confidence < 80) | (args.confidence > 100):
            msg = 'FATAL USER ERROR: Confdience must be between 80 and 100.'
            sys.exit(msg)

        if args.do_not_log_transform:
            args.log_transform = False
        else:
            args.log_transform = True
        
        args.confidence = args.confidence / 100

        if args.verbose:
            print_arguments(args)

        Command(args).test()

if __name__ == '__main__':

    AMiGA()



#!/usr/bin/env python

'''
AMiGA library for building kernels for Gaussian Process models.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (3 functions)

# buildKernel
# buildCompositeKernel
# addFixedKernel

import math
import numpy as np # type: ignore
import pandas as pd # type: ignore

from GPy.models import GPRegression # type: ignore
from GPy.kern import RBF,Fixed # type: ignore

from .config import config


def initHyperParams(ndim,x,y):
    '''
    Initializes hyper-parameters for RBF kernel based on method defined in libs/config.py
    Moment-based normalization sets lengthscale initial value to its range (max value - min value)
    and variance initial value is optimized with grid search centered around variance of observations.
    '''

    def oom(number): return math.floor(math.log(number,10))

    method = config['initialization_method']

    if method == 'default':
        return 1,1
    
    if method.startswith('moments-based'):
        
        lengthscale = np.max(x)-np.min(x)

        if method.endswith('fast') and ndim == 1:
            variance = np.var(y)
        elif method.endswith('proper') and ndim > 1:
            variance = np.var(y[:,0])

    if method.endswith('proper'):

        max_oom = 2 + oom(np.max(np.var(y)))
        lst_oom = [10**ii for ii in range(max_oom)]
        params = pd.DataFrame(lst_oom, columns=['variance'])

        loglikelihoods = []
        
        for _,(i_var) in params.iterrows():
            kern = RBF(1, ARD=False, variance=i_var, lengthscale=lengthscale)
            m = GPRegression(x,y,kern)
            m.optimize()
            loglikelihoods.append(m.objective_function())
        
        params = params.join(pd.DataFrame(data=loglikelihoods,columns=['LogLikelihood']))
        params = params.sort_values('LogLikelihood')
        variance = params.iloc[0,:].variance

    return variance, lengthscale


def buildKernel(ndim,x,y,ARD=False):

    variance, lengthscale = initHyperParams(ndim, x, y)

    if (ndim > 1) and (not ARD):
        kern = buildCompositeKernel(ndim,variance,lengthscale)

    elif (ndim > 1):
        kern = RBF(ndim,ARD=True,variance=variance,lengthscale=lengthscale)

    elif (ndim == 1):
        kern = RBF(ndim,ARD=False,variance=variance,lengthscale=lengthscale)

    return kern


def buildCompositeKernel(ndim,variance=1,lengthscale=1):
    '''
    Build an RBF kernel that extends beyond the first dimension. 
        For example, if GP(X,Y) and X has three dimensions, first is Time which 
        is a continous variable and 2nd and 3rd are categorical, then, 
        K = K1 * (K2+K3) = K1*K2 + K1*K3 where K1 will have lengthscale specific for time. 
        
        See also https://github.com/ptonner/gp_growth_phenotype/bgreat.py 
    '''

    kern = RBF(1,active_dims=[0],ARD=False,lengthscale=lengthscale,variance=variance)
    ksum = None

    for ii in range(1,ndim):
        if ksum is None:
            ksum = RBF(1,active_dims=[ii],ARD=False)
        else:
            ksum += RBF(1,active_dims=[ii],ARD=False)

    if ksum is None:
        return kern
    else:
        return kern*ksum


def addFixedKernel(kern,y_dim,error):
    '''
    Empirically estimate time-variant noise and model as a fixed kernel.
        This will results in an optimized GP model with low Gaussian noise hyperparameter, 
        but yet incoporates empirical noise in the optimization of hyperparameters.

    Args:
        kern (GPy.kern.src.rbf.RBF)
        ndim (int): number of dimensions of output variables, typically OD is 1-dimnsional
        error (numpy.ndarray): known variance (i.e. error) for each observation output variable 

    Ret:
        kern (GPy.kern.src.add.Add)
    '''

    cov = np.eye(len(error))*error # ndim x ndim 
    fixed = Fixed(y_dim,cov)
    kern = kern + fixed

    return kern


#!/usr/bin/env python

import subprocess

def test_help():
    result = subprocess.run(["amiga", "-h"], capture_output=True, text=True)
    assert "usage: amiga.py <command> [<args>]" in result.stdout.lower()

def test_print_defaults():
    result = subprocess.run(["amiga", "print_defaults"], capture_output=True, text=True)
    assert "default settings for select variables." in result.stdout.lower()

if __name__ == '__main__':
    test_help()
    test_print_defaults()



#!/usr/bin/env python

'''
AMiGA library of functions for printing and communicating via terminal.
'''

__author__ = "Firas S Midani"
__email__ = "midani@bcm.edu"


# TABLE OF CONTENTS (4 functions)

# smartPrint
# prettyNumberDisplay
# tidyDictPrint
# tidyMessage

import numpy as np # type: ignore


def smartPrint(msg,verbose):
    '''
    Only print if verbose argument is True. 

        Previously, I would concatenate messages inside a function and print once 
        function is completed, if verbose argument is satisfied. But the side 
        effect is that messages are printed in blocks (i.e. not flushing)
        and printing would not execute.

        In incremental printing (which this streamlines), if a function fails at a 
        specific point, this can be inferred based on where incremental printing would
        have been interrupted right after the point of failure.

    Args:
        msg (str)
        verbose (boolean)
    '''

    if verbose:
        print(msg)

def prettyNumberDisplay(num):
    '''
    Displays numebrs that are either large or small in scientific noation, otherwise does not.

    Args:
        num (float or int)
    '''

    if np.floor(np.abs(np.log10(np.abs(num)))) > 4:
        return f'{num:.3}'
    else:
        return f'{num:.3f}'


def tidyDictPrint(input_dict):
    '''
    Returns a message that neatly prints a dictionary into multiple lines. Each line
        is a key:value pair. Keys of dictionary are padded with period. Padding is 
        dynamically selected based on longest argument.
    '''

    if len(input_dict)==0:
        return 'None' 

    # dynamically set width of padding based on maximum argument length
    args = input_dict.keys()
    max_len = float(len(max(args,key=len))) # length of longest argument
    width = int(np.ceil(max_len/10)*10) # round up length to nearest ten

    # if width dos not add more than three padding spaces to argument, add 5 characters
    if (width - max_len) < 4:
        width += 5

    # compose multi-line message
    msg = ''
    for arg,value in input_dict.items():
        msg += '{:.<{width}}{}\n'.format(arg,value,width=width)

    return msg


def tidyMessage(msg):
    '''
    Returns a message in a text-based banner with header and footer
        composed of dashes and hashtags. Messae will be paddedd with flanking
        spaces. Length of message determines length of banner. Heigh of banner is 4 lines.

    Args:
        msg (str)

    Returns:
        msg_print (str)
    '''

    msg_len = len(msg)+2
    banner = '#{}#'.format('-'*msg_len)
    msg_print = f'{banner}\n# {msg} #\n{banner}\n'

    return msg_print

